import os

from contextlib import suppress
from enum import IntEnum, auto
from itertools import chain, count
from random import choice
from signal import SIGINT, SIGTERM, Signals
from signal import signal as signal_func
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple

from sanic.compat import OS_IS_WINDOWS
from sanic.exceptions import ServerKilled
from sanic.log import error_logger, logger
from sanic.worker.constants import RestartOrder
from sanic.worker.process import ProcessState, Worker, WorkerProcess
from sanic.worker.restarter import Restarter


if not OS_IS_WINDOWS:
    from signal import SIGKILL
else:
    SIGKILL = SIGINT


class MonitorCycle(IntEnum):
    BREAK = auto()
    CONTINUE = auto()


class WorkerManager:
    """Manage all of the processes.

    This class is used to manage all of the processes. It is instantiated
    by Sanic when in multiprocess mode (which is OOTB default) and is used
    to start, stop, and restart the worker processes.

    You can access it to interact with it **ONLY** when on the main process.

    Therefore, you should really only access it from within the
    `main_process_ready` event listener.

    ```python
    from sanic import Sanic

    app = Sanic("MyApp")

    @app.main_process_ready
    async def ready(app: Sanic, _):
        app.manager.manage("MyProcess", my_process, {"foo": "bar"})
    ```

    See [Worker Manager](/en/guide/deployment/manager) for more information.
    """

    THRESHOLD = WorkerProcess.THRESHOLD
    MAIN_NAME = "Sanic-Main"

    def __init__(
        self,
        number: int,
        serve,
        server_settings,
        context,
        monitor_pubsub,
        worker_state,
    ):
        self.num_server = number
        self.context = context
        self.transient: Dict[str, Worker] = {}
        self.durable: Dict[str, Worker] = {}
        self.restarter = Restarter()
        self.monitor_publisher, self.monitor_subscriber = monitor_pubsub
        self.worker_state = worker_state
        self.worker_state[self.MAIN_NAME] = {"pid": self.pid}
        self._shutting_down = False
        self._serve = serve
        self._server_settings = server_settings
        self._server_count = count()

        if number == 0:
            raise RuntimeError("Cannot serve with no workers")

        for _ in range(number):
            self.create_server()

        signal_func(SIGINT, self.shutdown_signal)
        signal_func(SIGTERM, self.shutdown_signal)

    def manage(
        self,
        name: str,
        func: Callable[..., Any],
        kwargs: Dict[str, Any],
        transient: bool = False,
        restartable: Optional[bool] = None,
        tracked: bool = True,
        auto_start: bool = True,
        workers: int = 1,
        ident: str = "",
    ) -> Worker:
        """Instruct Sanic to manage a custom process.

        Args:
            name (str): A name for the worker process
            func (Callable[..., Any]): The function to call in the background process
            kwargs (Dict[str, Any]): Arguments to pass to the function
            transient (bool, optional): Whether to mark the process as transient. If `True`
                then the Worker Manager will restart the process along
                with any global restart (ex: auto-reload), defaults to `False`
            restartable (Optional[bool], optional): Whether to mark the process as restartable. If
                `True` then the Worker Manager will be able to restart the process
                if prompted. If `transient=True`, this property will be implied
                to be `True`, defaults to `None`
            tracked (bool, optional): Whether to track the process after completion,
                defaults to `True`
            auto_start (bool, optional): Whether to start the process immediately, defaults to `True`
            workers (int, optional): The number of worker processes to run. Defaults to `1`.
            ident (str, optional): The identifier for the worker. If not provided, the name
                passed will be used. Defaults to `""`.

        Returns:
            Worker: The Worker instance
        """  # noqa: E501
        if name in self.transient or name in self.durable:
            raise ValueError(f"Worker {name} already exists")
        restartable = restartable if restartable is not None else transient
        if transient and not restartable:
            raise ValueError(
                "Cannot create a transient worker that is not restartable"
            )
        container = self.transient if transient else self.durable
        worker = Worker(
            ident or name,
            name,
            func,
            kwargs,
            self.context,
            self.worker_state,
            workers,
            restartable,
            tracked,
            auto_start,
        )
        container[worker.name] = worker
        return worker

    def create_server(self) -> Worker:
        """Create a new server process.

        Returns:
            Worker: The Worker instance
        """
        server_number = next(self._server_count)
        return self.manage(
            f"{WorkerProcess.SERVER_LABEL}-{server_number}",
            self._serve,
            self._server_settings,
            transient=True,
            restartable=True,
            ident=f"{WorkerProcess.SERVER_IDENTIFIER}{server_number:2}",
        )

    def shutdown_server(self, name: Optional[str] = None) -> None:
        """Shutdown a server process.

        Args:
            name (Optional[str], optional): The name of the server process to shutdown.
                If `None` then a random server will be chosen. Defaults to `None`.
        """  # noqa: E501
        if not name:
            servers = [
                worker
                for worker in self.transient.values()
                if worker.name.startswith(WorkerProcess.SERVER_LABEL)
            ]
            if not servers:
                error_logger.error(
                    "Server shutdown failed because a server was not found."
                )
                return
            worker = choice(servers)  # nosec B311
        else:
            worker = self.transient[name]

        for process in worker.processes:
            process.terminate()

        del self.transient[worker.name]

    def run(self):
        """Run the worker manager."""
        self.start()
        self.monitor()
        self.join()
        self.terminate()
        self.cleanup()

    def start(self):
        """Start the worker processes."""
        for worker in self.workers:
            for process in worker.processes:
                if not worker.auto_start:
                    process.set_state(ProcessState.NONE, True)
                    continue
                process.start()

    def join(self):
        """Join the worker processes."""
        logger.debug("Joining processes", extra={"verbosity": 1})
        joined = set()
        for process in self.processes:
            logger.debug(
                f"Found {process.pid} - {process.state.name}",
                extra={"verbosity": 1},
            )
            if process.state < ProcessState.JOINED:
                logger.debug(f"Joining {process.pid}", extra={"verbosity": 1})
                joined.add(process.pid)
                process.join()
        if joined:
            self.join()

    def terminate(self):
        """Terminate the worker processes."""
        if not self._shutting_down:
            for process in self.processes:
                process.terminate()

    def cleanup(self):
        """Cleanup the worker processes."""
        for process in self.processes:
            process.exit()

    def restart(
        self,
        process_names: Optional[List[str]] = None,
        restart_order=RestartOrder.SHUTDOWN_FIRST,
        **kwargs,
    ):
        """Restart the worker processes.

        Args:
            process_names (Optional[List[str]], optional): The names of the processes to restart.
                If `None` then all processes will be restarted. Defaults to `None`.
            restart_order (RestartOrder, optional): The order in which to restart the processes.
                Defaults to `RestartOrder.SHUTDOWN_FIRST`.
        """  # noqa: E501
        self.restarter.restart(
            transient_processes=list(self.transient_processes),
            durable_processes=list(self.durable_processes),
            process_names=process_names,
            restart_order=restart_order,
            **kwargs,
        )

    def scale(self, num_worker: int):
        if num_worker <= 0:
            raise ValueError("Cannot scale to 0 workers.")

        change = num_worker - self.num_server
        if change == 0:
            logger.info(
                f"No change needed. There are already {num_worker} workers."
            )
            return

        logger.info(f"Scaling from {self.num_server} to {num_worker} workers")
        if change > 0:
            for _ in range(change):
                worker = self.create_server()
                for process in worker.processes:
                    process.start()
        else:
            for _ in range(abs(change)):
                self.shutdown_server()
        self.num_server = num_worker

    def monitor(self):
        """Monitor the worker processes.

        First, wait for all of the workers to acknowledge that they are ready.
        Then, wait for messages from the workers. If a message is received
        then it is processed and the state of the worker is updated.

        Also used to restart, shutdown, and scale the workers.

        Raises:
            ServerKilled: Raised when a worker fails to come online.
        """
        self.wait_for_ack()
        while True:
            try:
                cycle = self._poll_monitor()
                if cycle is MonitorCycle.BREAK:
                    break
                elif cycle is MonitorCycle.CONTINUE:
                    continue
                self._sync_states()
                self._cleanup_non_tracked_workers()
            except InterruptedError:
                if not OS_IS_WINDOWS:
                    raise
                break

    def wait_for_ack(self):  # no cov
        """Wait for all of the workers to acknowledge that they are ready."""
        misses = 0
        message = (
            "It seems that one or more of your workers failed to come "
            "online in the allowed time. Sanic is shutting down to avoid a "
            f"deadlock. The current threshold is {self.THRESHOLD / 10}s. "
            "If this problem persists, please check out the documentation "
            "https://sanic.dev/en/guide/deployment/manager.html#worker-ack."
        )
        while not self._all_workers_ack():
            if self.monitor_subscriber.poll(0.1):
                monitor_msg = self.monitor_subscriber.recv()
                if monitor_msg != "__TERMINATE_EARLY__":
                    self.monitor_publisher.send(monitor_msg)
                    continue
                misses = self.THRESHOLD
                message = (
                    "One of your worker processes terminated before startup "
                    "was completed. Please solve any errors experienced "
                    "during startup. If you do not see an exception traceback "
                    "in your error logs, try running Sanic in a single "
                    "process using --single-process or single_process=True. "
                    "Once you are confident that the server is able to start "
                    "without errors you can switch back to multiprocess mode."
                )
            misses += 1
            if misses > self.THRESHOLD:
                error_logger.error(
                    "Not all workers acknowledged a successful startup. "
                    "Shutting down.\n\n" + message
                )
                self.kill()

    @property
    def workers(self) -> List[Worker]:
        """Get all of the workers."""
        return list(self.transient.values()) + list(self.durable.values())

    @property
    def all_workers(self) -> Iterable[Tuple[str, Worker]]:
        return chain(self.transient.items(), self.durable.items())

    @property
    def processes(self):
        """Get all of the processes."""
        for worker in self.workers:
            for process in worker.processes:
                if not process.pid:
                    continue
                yield process

    @property
    def transient_processes(self):
        """Get all of the transient processes."""
        for worker in self.transient.values():
            for process in worker.processes:
                yield process

    @property
    def durable_processes(self):
        for worker in self.durable.values():
            for process in worker.processes:
                yield process

    def kill(self):
        """Kill all of the processes."""
        for process in self.processes:
            logger.info("Killing %s [%s]", process.name, process.pid)
            with suppress(ProcessLookupError):
                os.kill(process.pid, SIGKILL)
        raise ServerKilled

    def shutdown_signal(self, signal, frame):
        """Handle the shutdown signal."""
        if self._shutting_down:
            logger.info("Shutdown interrupted. Killing.")
            with suppress(ServerKilled):
                self.kill()

        logger.info("Received signal %s. Shutting down.", Signals(signal).name)
        self.monitor_publisher.send(None)
        self.shutdown()

    def shutdown(self):
        """Shutdown the worker manager."""
        for process in self.processes:
            if process.is_alive():
                process.terminate()
        self._shutting_down = True

    def remove_worker(self, worker: Worker) -> None:
        if worker.tracked:
            error_logger.error(
                f"Worker {worker.name} is tracked and cannot be removed."
            )
            return
        if worker.has_alive_processes():
            error_logger.error(
                f"Worker {worker.name} has alive processes and cannot be "
                "removed."
            )
            return
        self.transient.pop(worker.name, None)
        self.durable.pop(worker.name, None)
        for process in worker.processes:
            self.worker_state.pop(process.name, None)
        logger.info("Removed worker %s", worker.name)
        del worker

    @property
    def pid(self):
        """Get the process ID of the main process."""
        return os.getpid()

    def _all_workers_ack(self):
        acked = [
            worker_state.get("state") == ProcessState.ACKED.name
            for worker_state in self.worker_state.values()
            if worker_state.get("server")
        ]
        return all(acked) and len(acked) == self.num_server

    def _sync_states(self):
        for process in self.processes:
            try:
                state = self.worker_state[process.name].get("state")
            except KeyError:
                process.set_state(ProcessState.TERMINATED, True)
                continue
            if not process.is_alive():
                state = "FAILED" if process.exitcode else "COMPLETED"
            if state and process.state.name != state:
                process.set_state(ProcessState[state], True)

    def _cleanup_non_tracked_workers(self) -> None:
        to_remove = [
            worker
            for worker in self.workers
            if not worker.tracked and not worker.has_alive_processes()
        ]

        for worker in to_remove:
            self.remove_worker(worker)

    def _poll_monitor(self) -> Optional[MonitorCycle]:
        if self.monitor_subscriber.poll(0.1):
            message = self.monitor_subscriber.recv()
            logger.debug(f"Monitor message: {message}", extra={"verbosity": 2})
            if not message:
                return MonitorCycle.BREAK
            elif message == "__TERMINATE__":
                self._handle_terminate()
                return MonitorCycle.BREAK
            elif isinstance(message, tuple) and (
                len(message) == 7 or len(message) == 8
            ):
                self._handle_manage(*message)  # type: ignore
                return MonitorCycle.CONTINUE
            elif not isinstance(message, str):
                error_logger.error(
                    "Monitor received an invalid message: %s", message
                )
                return MonitorCycle.CONTINUE
            return self._handle_message(message)
        return None

    def _handle_terminate(self) -> None:
        self.shutdown()

    def _handle_message(self, message: str) -> Optional[MonitorCycle]:
        logger.debug(
            "Incoming monitor message: %s",
            message,
            extra={"verbosity": 1},
        )
        split_message = message.split(":", 2)
        if message.startswith("__SCALE__"):
            self.scale(int(split_message[-1]))
            return MonitorCycle.CONTINUE

        processes = split_message[0]
        reloaded_files = split_message[1] if len(split_message) > 1 else None
        process_names: Optional[List[str]] = [
            name.strip() for name in processes.split(",")
        ]
        if process_names and "__ALL_PROCESSES__" in process_names:
            process_names = None
        order = (
            RestartOrder.STARTUP_FIRST
            if "STARTUP_FIRST" in split_message
            else RestartOrder.SHUTDOWN_FIRST
        )
        self.restart(
            process_names=process_names,
            reloaded_files=reloaded_files,
            restart_order=order,
        )

        return None

    def _handle_manage(
        self,
        name: str,
        func: Callable[..., Any],
        kwargs: Dict[str, Any],
        transient: bool,
        restartable: Optional[bool],
        tracked: bool,
        auto_start: bool,
        workers: int,
    ) -> None:
        try:
            worker = self.manage(
                name,
                func,
                kwargs,
                transient=transient,
                restartable=restartable,
                tracked=tracked,
                auto_start=auto_start,
                workers=workers,
            )
        except Exception:
            error_logger.exception("Failed to manage worker %s", name)
        else:
            if not auto_start:
                return
            for process in worker.processes:
                process.start()



import concurrent.futures
import sys
import threading
import time
import traceback

from ast import PyCF_ALLOW_TOP_LEVEL_AWAIT
from asyncio import iscoroutine, new_event_loop
from code import InteractiveConsole
from types import FunctionType
from typing import Any, Dict, NamedTuple, Optional, Sequence, Tuple, Union

import sanic

from sanic import Request, Sanic
from sanic.compat import Header
from sanic.helpers import Default
from sanic.http.constants import Stage
from sanic.log import Colors
from sanic.models.protocol_types import TransportProtocol
from sanic.response.types import HTTPResponse


try:
    from httpx import Client

    HTTPX_AVAILABLE = True

    class SanicClient(Client):
        def __init__(self, app: Sanic):
            base_url = app.get_server_location(
                app.state.server_info[0].settings
            )
            super().__init__(base_url=base_url)

except ImportError:
    HTTPX_AVAILABLE = False

try:
    import readline  # noqa
except ImportError:
    print(
        "Module 'readline' not available. History navigation will be limited.",
        file=sys.stderr,
    )

repl_app: Optional[Sanic] = None
repl_response: Optional[HTTPResponse] = None


class REPLProtocol(TransportProtocol):
    def __init__(self):
        self.stage = Stage.IDLE
        self.request_body = True

    def respond(self, response):
        global repl_response
        repl_response = response
        response.stream = self
        return response

    async def send(self, data, end_stream): ...


class Result(NamedTuple):
    request: Request
    response: HTTPResponse


def make_request(
    url: str = "/",
    headers: Optional[Union[Dict[str, Any], Sequence[Tuple[str, str]]]] = None,
    method: str = "GET",
    body: Optional[str] = None,
):
    assert repl_app, "No Sanic app has been registered."
    headers = headers or {}
    protocol = REPLProtocol()
    request = Request(  # type: ignore
        url.encode(),
        Header(headers),
        "1.1",
        method,
        protocol,
        repl_app,
    )
    if body is not None:
        request.body = body.encode()
    request.stream = protocol  # type: ignore
    request.conn_info = None
    return request


async def respond(request) -> HTTPResponse:
    assert repl_app, "No Sanic app has been registered."
    await repl_app.handle_request(request)
    assert repl_response
    return repl_response


async def do(
    url: str = "/",
    headers: Optional[Union[Dict[str, Any], Sequence[Tuple[str, str]]]] = None,
    method: str = "GET",
    body: Optional[str] = None,
) -> Result:
    request = make_request(url, headers, method, body)
    response = await respond(request)
    return Result(request, response)


class SanicREPL(InteractiveConsole):
    def __init__(self, app: Sanic, start: Optional[Default] = None):
        global repl_app
        repl_app = app
        locals_available = {
            "app": app,
            "sanic": sanic,
            "do": do,
        }
        client_availability = ""
        variable_descriptions = [
            f"  - {Colors.BOLD + Colors.SANIC}app{Colors.END}: The Sanic application instance - {Colors.BOLD + Colors.BLUE}{str(app)}{Colors.END}",  # noqa: E501
            f"  - {Colors.BOLD + Colors.SANIC}sanic{Colors.END}: The Sanic module - {Colors.BOLD + Colors.BLUE}import sanic{Colors.END}",  # noqa: E501
            f"  - {Colors.BOLD + Colors.SANIC}do{Colors.END}: An async function to fake a request to the application - {Colors.BOLD + Colors.BLUE}Result(request, response){Colors.END}",  # noqa: E501
        ]
        if HTTPX_AVAILABLE:
            locals_available["client"] = SanicClient(app)
            variable_descriptions.append(
                f"  - {Colors.BOLD + Colors.SANIC}client{Colors.END}: A client to access the Sanic app instance using httpx - {Colors.BOLD + Colors.BLUE}from httpx import Client{Colors.END}",  # noqa: E501
            )
        else:
            client_availability = (
                f"\n{Colors.YELLOW}The HTTP client has been disabled. "
                "To enable it, install httpx:\n\t"
                f"pip install httpx{Colors.END}\n"
            )
        super().__init__(locals=locals_available)
        self.compile.compiler.flags |= PyCF_ALLOW_TOP_LEVEL_AWAIT
        self.loop = new_event_loop()
        self._start = start
        self._pause_event = threading.Event()
        self._started_event = threading.Event()
        self._interact_thread = threading.Thread(
            target=self._console,
            daemon=True,
        )
        self._monitor_thread = threading.Thread(
            target=self._monitor,
            daemon=True,
        )
        self._async_thread = threading.Thread(
            target=self.loop.run_forever,
            daemon=True,
        )
        self.app = app
        self.resume()
        self.exit_message = "Closing the REPL."
        self.banner_message = "\n".join(
            [
                f"\n{Colors.BOLD}Welcome to the Sanic interactive console{Colors.END}",  # noqa: E501
                client_availability,
                "The following objects are available for your convenience:",  # noqa: E501
                *variable_descriptions,
                "\nThe async/await keywords are available for use here.",  # noqa: E501
                f"To exit, press {Colors.BOLD}CTRL+C{Colors.END}, "
                f"{Colors.BOLD}CTRL+D{Colors.END}, or type {Colors.BOLD}exit(){Colors.END}.\n",  # noqa: E501
            ]
        )

    def pause(self):
        if self.is_paused():
            return
        self._pause_event.clear()

    def resume(self):
        self._pause_event.set()

    def runsource(self, source, filename="<input>", symbol="single"):
        if source.strip() == "exit()":
            self._shutdown()
            return False

        if self.is_paused():
            print("Console is paused. Please wait for it to be resumed.")
            return False

        return super().runsource(source, filename, symbol)

    def runcode(self, code):
        future = concurrent.futures.Future()

        async def callback():
            func = FunctionType(code, self.locals)
            try:
                result = func()
                if iscoroutine(result):
                    result = await result
            except BaseException:
                traceback.print_exc()
                result = False
            future.set_result(result)

        self.loop.call_soon_threadsafe(self.loop.create_task, callback())
        return future.result()

    def is_paused(self):
        return not self._pause_event.is_set()

    def _console(self):
        self._started_event.set()
        self.interact(banner=self.banner_message, exitmsg=self.exit_message)
        self._shutdown()

    def _monitor(self):
        if isinstance(self._start, Default):
            enter = f"{Colors.BOLD + Colors.SANIC}ENTER{Colors.END}"
            start = input(f"\nPress {enter} at anytime to start the REPL.\n\n")
            if start:
                return
        try:
            while True:
                if not self._started_event.is_set():
                    self.app.manager.wait_for_ack()
                    self._interact_thread.start()
                elif self.app.manager._all_workers_ack() and self.is_paused():
                    self.resume()
                    print(sys.ps1, end="", flush=True)
                elif (
                    not self.app.manager._all_workers_ack()
                    and not self.is_paused()
                ):
                    self.pause()
                time.sleep(0.1)
        except (ConnectionResetError, BrokenPipeError):
            pass

    def _shutdown(self):
        self.app.manager.monitor_publisher.send("__TERMINATE__")

    def run(self):
        self._monitor_thread.start()
        self._async_thread.start()



from __future__ import annotations

import re
import string
import sys

from datetime import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from sanic.exceptions import ServerError
from sanic.log import deprecation


if TYPE_CHECKING:
    from sanic.compat import Header

if sys.version_info < (3, 8):  # no cov
    SameSite = str
else:  # no cov
    from typing import Literal

    SameSite = Union[
        Literal["Strict"],
        Literal["Lax"],
        Literal["None"],
        Literal["strict"],
        Literal["lax"],
        Literal["none"],
    ]

DEFAULT_MAX_AGE = 0
SAMESITE_VALUES = ("strict", "lax", "none")

LEGAL_CHARS = string.ascii_letters + string.digits + "!#$%&'*+-.^_`|~:"
UNESCAPED_CHARS = LEGAL_CHARS + " ()/<=>?@[]{}"
TRANSLATOR = {ch: f"\\{ch:03o}" for ch in bytes(range(32)) + b'";\\\x7f'}


def _quote(str):  # no cov
    r"""Quote a string for use in a cookie header.
    If the string does not need to be double-quoted, then just return the
    string.  Otherwise, surround the string in doublequotes and quote
    (with a \) special characters.
    """
    if str is None or _is_legal_key(str):
        return str
    else:
        return f'"{str.translate(TRANSLATOR)}"'


_is_legal_key = re.compile("[%s]+" % re.escape(LEGAL_CHARS)).fullmatch


# In v24.9, we should remove this as being a subclass of dict
class CookieJar(dict):
    """A container to manipulate cookies.

    CookieJar dynamically writes headers as cookies are added and removed
    It gets around the limitation of one header per name by using the
    MultiHeader class to provide a unique key that encodes to Set-Cookie.

    Args:
        headers (Header): The headers object to write cookies to.
    """

    HEADER_KEY = "Set-Cookie"

    def __init__(self, headers: Header):
        super().__init__()
        self.headers = headers

    def __setitem__(self, key, value):
        # If this cookie doesn't exist, add it to the header keys
        deprecation(
            "Setting cookie values using the dict pattern has been "
            "deprecated. You should instead use the cookies.add_cookie "
            "method. To learn more, please see: "
            "https://sanic.dev/en/guide/release-notes/v23.3.html#response-cookies",  # noqa
            0,
        )
        if key not in self:
            self.add_cookie(key, value, secure=False, samesite=None)
        else:
            self[key].value = value

    def __delitem__(self, key):
        deprecation(
            "Deleting cookie values using the dict pattern has been "
            "deprecated. You should instead use the cookies.delete_cookie "
            "method. To learn more, please see: "
            "https://sanic.dev/en/guide/release-notes/v23.3.html#response-cookies",  # noqa
            0,
        )
        if key in self:
            super().__delitem__(key)
        self.delete_cookie(key)

    def __len__(self):  # no cov
        return len(self.cookies)

    def __getitem__(self, key: str) -> Cookie:
        deprecation(
            "Accessing cookies from the CookieJar by dict key is deprecated. "
            "You should instead use the cookies.get_cookie method. "
            "To learn more, please see: "
            "https://sanic.dev/en/guide/release-notes/v23.3.html#response-cookies",  # noqa
            0,
        )
        return super().__getitem__(key)

    def __iter__(self):  # no cov
        deprecation(
            "Iterating over the CookieJar has been deprecated and will be "
            "removed in v24.9. To learn more, please see: "
            "https://sanic.dev/en/guide/release-notes/v23.3.html#response-cookies",  # noqa
            24.9,
        )
        return super().__iter__()

    def keys(self):  # no cov
        """Deprecated in v24.9"""
        deprecation(
            "Accessing CookieJar.keys() has been deprecated and will be "
            "removed in v24.9. To learn more, please see: "
            "https://sanic.dev/en/guide/release-notes/v23.3.html#response-cookies",  # noqa
            24.9,
        )
        return super().keys()

    def values(self):  # no cov
        """Deprecated in v24.9"""
        deprecation(
            "Accessing CookieJar.values() has been deprecated and will be "
            "removed in v24.9. To learn more, please see: "
            "https://sanic.dev/en/guide/release-notes/v23.3.html#response-cookies",  # noqa
            24.9,
        )
        return super().values()

    def items(self):  # no cov
        """Deprecated in v24.9"""
        deprecation(
            "Accessing CookieJar.items() has been deprecated and will be "
            "removed in v24.9. To learn more, please see: "
            "https://sanic.dev/en/guide/release-notes/v23.3.html#response-cookies",  # noqa
            24.9,
        )
        return super().items()

    def get(self, *args, **kwargs):  # no cov
        """Deprecated in v24.9"""
        deprecation(
            "Accessing cookies from the CookieJar using get is deprecated "
            "and will be removed in v24.9. You should instead use the "
            "cookies.get_cookie method. To learn more, please see: "
            "https://sanic.dev/en/guide/release-notes/v23.3.html#response-cookies",  # noqa
            24.9,
        )
        return super().get(*args, **kwargs)

    def pop(self, key, *args, **kwargs):  # no cov
        """Deprecated in v24.9"""
        deprecation(
            "Using CookieJar.pop() has been deprecated and will be "
            "removed in v24.9. To learn more, please see: "
            "https://sanic.dev/en/guide/release-notes/v23.3.html#response-cookies",  # noqa
            24.9,
        )
        self.delete(key)
        return super().pop(key, *args, **kwargs)

    @property
    def header_key(self):  # no cov
        """Deprecated in v24.9"""
        deprecation(
            "The CookieJar.header_key property has been deprecated and will "
            "be removed in version 24.9. Use CookieJar.HEADER_KEY. ",
            24.9,
        )
        return CookieJar.HEADER_KEY

    @property
    def cookie_headers(self) -> Dict[str, str]:  # no cov
        """Deprecated in v24.9"""
        deprecation(
            "The CookieJar.coookie_headers property has been deprecated "
            "and will be removed in version 24.9. If you need to check if a "
            "particular cookie key has been set, use CookieJar.has_cookie.",
            24.9,
        )
        return {key: self.header_key for key in self}

    @property
    def cookies(self) -> List[Cookie]:
        """A list of cookies in the CookieJar.

        Returns:
            List[Cookie]: A list of cookies in the CookieJar.
        """
        return self.headers.getall(self.HEADER_KEY)

    def get_cookie(
        self,
        key: str,
        path: str = "/",
        domain: Optional[str] = None,
        host_prefix: bool = False,
        secure_prefix: bool = False,
    ) -> Optional[Cookie]:
        """Fetch a cookie from the CookieJar.

        Args:
            key (str): The key of the cookie to fetch.
            path (str, optional): The path of the cookie. Defaults to `"/"`.
            domain (Optional[str], optional): The domain of the cookie.
                Defaults to `None`.
            host_prefix (bool, optional): Whether to add __Host- as a prefix to the key.
                This requires that path="/", domain=None, and secure=True.
                Defaults to `False`.
            secure_prefix (bool, optional): Whether to add __Secure- as a prefix to the key.
                This requires that secure=True. Defaults to `False`.

        Returns:
            Optional[Cookie]: The cookie if it exists, otherwise `None`.
        """  # noqa: E501
        for cookie in self.cookies:
            if (
                cookie.key == Cookie.make_key(key, host_prefix, secure_prefix)
                and cookie.path == path
                and cookie.domain == domain
            ):
                return cookie
        return None

    def has_cookie(
        self,
        key: str,
        path: str = "/",
        domain: Optional[str] = None,
        host_prefix: bool = False,
        secure_prefix: bool = False,
    ) -> bool:
        """Check if a cookie exists in the CookieJar.

        Args:
            key (str): The key of the cookie to check.
            path (str, optional): The path of the cookie. Defaults to `"/"`.
            domain (Optional[str], optional): The domain of the cookie.
                Defaults to `None`.
            host_prefix (bool, optional): Whether to add __Host- as a prefix to the key.
                This requires that path="/", domain=None, and secure=True.
                Defaults to `False`.
            secure_prefix (bool, optional): Whether to add __Secure- as a prefix to the key.
                This requires that secure=True. Defaults to `False`.

        Returns:
            bool: Whether the cookie exists.
        """  # noqa: E501
        for cookie in self.cookies:
            if (
                cookie.key == Cookie.make_key(key, host_prefix, secure_prefix)
                and cookie.path == path
                and cookie.domain == domain
            ):
                return True
        return False

    def add_cookie(
        self,
        key: str,
        value: str,
        *,
        path: str = "/",
        domain: Optional[str] = None,
        secure: bool = True,
        max_age: Optional[int] = None,
        expires: Optional[datetime] = None,
        httponly: bool = False,
        samesite: Optional[SameSite] = "Lax",
        partitioned: bool = False,
        comment: Optional[str] = None,
        host_prefix: bool = False,
        secure_prefix: bool = False,
    ) -> Cookie:
        """Add a cookie to the CookieJar.

        Args:
            key (str): Key of the cookie.
            value (str): Value of the cookie.
            path (str, optional): Path of the cookie. Defaults to "/".
            domain (Optional[str], optional): Domain of the cookie. Defaults to None.
            secure (bool, optional): Whether to set it as a secure cookie. Defaults to True.
            max_age (Optional[int], optional): Max age of the cookie in seconds; if set to 0 a
                browser should delete it. Defaults to None.
            expires (Optional[datetime], optional): When the cookie expires; if set to None browsers
                should set it as a session cookie. Defaults to None.
            httponly (bool, optional): Whether to set it as HTTP only. Defaults to False.
            samesite (Optional[SameSite], optional): How to set the samesite property, should be
                strict, lax, or none (case insensitive). Defaults to "Lax".
            partitioned (bool, optional): Whether to set it as partitioned. Defaults to False.
            comment (Optional[str], optional): A cookie comment. Defaults to None.
            host_prefix (bool, optional): Whether to add __Host- as a prefix to the key.
                This requires that path="/", domain=None, and secure=True. Defaults to False.
            secure_prefix (bool, optional): Whether to add __Secure- as a prefix to the key.
                This requires that secure=True. Defaults to False.

        Returns:
            Cookie: The instance of the created cookie.

        Raises:
            ServerError: If host_prefix is set without secure=True.
            ServerError: If host_prefix is set without path="/" and domain=None.
            ServerError: If host_prefix is set with domain.
            ServerError: If secure_prefix is set without secure=True.
            ServerError: If partitioned is set without host_prefix=True.

        Examples:
            Basic usage
            ```python
            cookie = add_cookie('name', 'value')
            ```

            Adding a cookie with a custom path and domain
            ```python
            cookie = add_cookie('name', 'value', path='/custom', domain='example.com')
            ```

            Adding a secure, HTTP-only cookie with a comment
            ```python
            cookie = add_cookie('name', 'value', secure=True, httponly=True, comment='My Cookie')
            ```

            Adding a cookie with a max age of 60 seconds
            ```python
            cookie = add_cookie('name', 'value', max_age=60)
            ```
        """  # noqa: E501
        cookie = Cookie(
            key,
            value,
            path=path,
            expires=expires,
            comment=comment,
            domain=domain,
            max_age=max_age,
            secure=secure,
            httponly=httponly,
            samesite=samesite,
            partitioned=partitioned,
            host_prefix=host_prefix,
            secure_prefix=secure_prefix,
        )
        self.headers.add(self.HEADER_KEY, cookie)

        # This should be removed in v24.9
        super().__setitem__(key, cookie)

        return cookie

    def delete_cookie(
        self,
        key: str,
        *,
        path: str = "/",
        domain: Optional[str] = None,
        secure: bool = True,
        host_prefix: bool = False,
        secure_prefix: bool = False,
    ) -> None:
        """
        Delete a cookie

        This will effectively set it as Max-Age: 0, which a browser should
        interpret it to mean: "delete the cookie".

        Since it is a browser/client implementation, your results may vary
        depending upon which client is being used.

        :param key: The key to be deleted
        :type key: str
        :param path: Path of the cookie, defaults to None
        :type path: Optional[str], optional
        :param domain: Domain of the cookie, defaults to None
        :type domain: Optional[str], optional
        :param secure: Whether to delete a secure cookie. Defaults to True.
        :param secure: bool
        :param host_prefix: Whether to add __Host- as a prefix to the key.
            This requires that path="/", domain=None, and secure=True,
            defaults to False
        :type host_prefix: bool
        :param secure_prefix: Whether to add __Secure- as a prefix to the key.
            This requires that secure=True, defaults to False
        :type secure_prefix: bool
        """
        if host_prefix and not (secure and path == "/" and domain is None):
            raise ServerError(
                "Cannot set host_prefix on a cookie without "
                "path='/', domain=None, and secure=True"
            )
        if secure_prefix and not secure:
            raise ServerError(
                "Cannot set secure_prefix on a cookie without secure=True"
            )

        cookies: List[Cookie] = self.headers.popall(self.HEADER_KEY, [])
        existing_cookie = None
        for cookie in cookies:
            if (
                cookie.key != Cookie.make_key(key, host_prefix, secure_prefix)
                or cookie.path != path
                or cookie.domain != domain
            ):
                self.headers.add(self.HEADER_KEY, cookie)
            elif existing_cookie is None:
                existing_cookie = cookie
        # This should be removed in v24.9
        try:
            super().__delitem__(key)
        except KeyError:
            ...

        if existing_cookie is not None:
            # Use all the same values as the cookie to be deleted
            # except value="" and max_age=0
            self.add_cookie(
                key=key,
                value="",
                path=existing_cookie.path,
                domain=existing_cookie.domain,
                secure=existing_cookie.secure,
                max_age=0,
                httponly=existing_cookie.httponly,
                partitioned=existing_cookie.partitioned,
                samesite=existing_cookie.samesite,
                host_prefix=host_prefix,
                secure_prefix=secure_prefix,
            )
        else:
            self.add_cookie(
                key=key,
                value="",
                path=path,
                domain=domain,
                secure=secure,
                max_age=0,
                samesite=None,
                host_prefix=host_prefix,
                secure_prefix=secure_prefix,
            )


# In v24.9, we should remove this as being a subclass of dict
# Instead, it should be an object with __slots__
# All of the current property accessors should be removed in favor
# of actual slotted properties.
class Cookie(dict):
    """A representation of a HTTP cookie, providing an interface to manipulate cookie attributes intended for a response.

    This class is a simplified representation of a cookie, similar to the Morsel SimpleCookie in Python's standard library.
    It allows the manipulation of various cookie attributes including path, domain, security settings, and others.

    Several "smart defaults" are provided to make it easier to create cookies that are secure by default. These include:

    - Setting the `secure` flag to `True` by default
    - Setting the `samesite` flag to `Lax` by default

    Args:
        key (str): The key (name) of the cookie.
        value (str): The value of the cookie.
        path (str, optional): The path for the cookie. Defaults to "/".
        domain (Optional[str], optional): The domain for the cookie.
            Defaults to `None`.
        secure (bool, optional): Whether the cookie is secure.
            Defaults to `True`.
        max_age (Optional[int], optional): The maximum age of the cookie
            in seconds. Defaults to `None`.
        expires (Optional[datetime], optional): The expiration date of the
            cookie. Defaults to `None`.
        httponly (bool, optional): HttpOnly flag for the cookie.
            Defaults to `False`.
        samesite (Optional[SameSite], optional): The SameSite attribute for
            the cookie. Defaults to `"Lax"`.
        partitioned (bool, optional): Whether the cookie is partitioned.
            Defaults to `False`.
        comment (Optional[str], optional): A comment for the cookie.
            Defaults to `None`.
        host_prefix (bool, optional): Whether to use the host prefix.
            Defaults to `False`.
        secure_prefix (bool, optional): Whether to use the secure prefix.
            Defaults to `False`.
    """  # noqa: E501

    HOST_PREFIX = "__Host-"
    SECURE_PREFIX = "__Secure-"

    _keys = {
        "path": "Path",
        "comment": "Comment",
        "domain": "Domain",
        "max-age": "Max-Age",
        "expires": "expires",
        "samesite": "SameSite",
        "version": "Version",
        "secure": "Secure",
        "httponly": "HttpOnly",
        "partitioned": "Partitioned",
    }
    _flags = {"secure", "httponly", "partitioned"}

    def __init__(
        self,
        key: str,
        value: str,
        *,
        path: str = "/",
        domain: Optional[str] = None,
        secure: bool = True,
        max_age: Optional[int] = None,
        expires: Optional[datetime] = None,
        httponly: bool = False,
        samesite: Optional[SameSite] = "Lax",
        partitioned: bool = False,
        comment: Optional[str] = None,
        host_prefix: bool = False,
        secure_prefix: bool = False,
    ):
        if key in self._keys:
            raise KeyError("Cookie name is a reserved word")
        if not _is_legal_key(key):
            raise KeyError("Cookie key contains illegal characters")
        if host_prefix:
            if not secure:
                raise ServerError(
                    "Cannot set host_prefix on a cookie without secure=True"
                )
            if path != "/":
                raise ServerError(
                    "Cannot set host_prefix on a cookie unless path='/'"
                )
            if domain:
                raise ServerError(
                    "Cannot set host_prefix on a cookie with a defined domain"
                )
        elif secure_prefix and not secure:
            raise ServerError(
                "Cannot set secure_prefix on a cookie without secure=True"
            )
        if partitioned and not host_prefix:
            # This is technically possible, but it is not advisable so we will
            # take a stand and say "don't shoot yourself in the foot"
            raise ServerError(
                "Cannot create a partitioned cookie without "
                "also setting host_prefix=True"
            )

        self.key = self.make_key(key, host_prefix, secure_prefix)
        self.value = value
        super().__init__()

        # This is a temporary solution while this object is a dict. We update
        # all of the values in bulk, except for the values that have
        # key-specific validation in _set_value
        self.update(
            {
                "path": path,
                "comment": comment,
                "domain": domain,
                "secure": secure,
                "httponly": httponly,
                "partitioned": partitioned,
                "expires": None,
                "max-age": None,
                "samesite": None,
            }
        )
        if expires is not None:
            self._set_value("expires", expires)
        if max_age is not None:
            self._set_value("max-age", max_age)
        if samesite is not None:
            self._set_value("samesite", samesite)

    def __setitem__(self, key, value):
        deprecation(
            "Setting values on a Cookie object as a dict has been deprecated. "
            "This feature will be removed in v24.9. You should instead set "
            f"values on cookies as object properties: cookie.{key}=... ",
            24.9,
        )
        self._set_value(key, value)

    # This is a temporary method for backwards compat and should be removed
    # in v24.9 when this is no longer a dict
    def _set_value(self, key: str, value: Any) -> None:
        if key not in self._keys:
            raise KeyError("Unknown cookie property: %s=%s" % (key, value))

        if value is not None:
            if key.lower() == "max-age" and not str(value).isdigit():
                raise ValueError("Cookie max-age must be an integer")
            elif key.lower() == "expires" and not isinstance(value, datetime):
                raise TypeError("Cookie 'expires' property must be a datetime")
            elif key.lower() == "samesite":
                if value.lower() not in SAMESITE_VALUES:
                    raise TypeError(
                        "Cookie 'samesite' property must "
                        f"be one of: {','.join(SAMESITE_VALUES)}"
                    )
                value = value.title()

        super().__setitem__(key, value)

    def encode(self, encoding: str) -> bytes:
        """Encode the cookie content in a specific type of encoding instructed by the developer.

        Leverages the `str.encode` method provided by Python.

        This method can be used to encode and embed ``utf-8`` content into
        the cookies.

        .. warning::
            Direct encoding of a Cookie object has been deprecated and will be removed in v24.9.

        Args:
            encoding (str): The encoding type to be used.

        Returns:
            bytes: The encoded cookie content.
        """  # noqa: E501
        deprecation(
            "Direct encoding of a Cookie object has been deprecated and will "
            "be removed in v24.9.",
            24.9,
        )
        return str(self).encode(encoding)

    def __str__(self):
        """Format as a Set-Cookie header value."""
        output = ["%s=%s" % (self.key, _quote(self.value))]
        key_index = list(self._keys)
        for key, value in sorted(
            self.items(), key=lambda x: key_index.index(x[0])
        ):
            if value is not None and value is not False:
                if key == "max-age":
                    try:
                        output.append("%s=%d" % (self._keys[key], value))
                    except TypeError:
                        output.append("%s=%s" % (self._keys[key], value))
                elif key == "expires":
                    output.append(
                        "%s=%s"
                        % (
                            self._keys[key],
                            value.strftime("%a, %d-%b-%Y %T GMT"),
                        )
                    )
                elif key in self._flags:
                    output.append(self._keys[key])
                else:
                    output.append("%s=%s" % (self._keys[key], value))

        return "; ".join(output)

    @property
    def path(self) -> str:  # no cov
        """The path of the cookie. Defaults to `"/"`."""
        return self["path"]

    @path.setter
    def path(self, value: str) -> None:  # no cov
        self._set_value("path", value)

    @property
    def expires(self) -> Optional[datetime]:  # no cov
        """The expiration date of the cookie. Defaults to `None`."""
        return self.get("expires")

    @expires.setter
    def expires(self, value: datetime) -> None:  # no cov
        self._set_value("expires", value)

    @property
    def comment(self) -> Optional[str]:  # no cov
        """A comment for the cookie. Defaults to `None`."""
        return self.get("comment")

    @comment.setter
    def comment(self, value: str) -> None:  # no cov
        self._set_value("comment", value)

    @property
    def domain(self) -> Optional[str]:  # no cov
        """The domain of the cookie. Defaults to `None`."""
        return self.get("domain")

    @domain.setter
    def domain(self, value: str) -> None:  # no cov
        self._set_value("domain", value)

    @property
    def max_age(self) -> Optional[int]:  # no cov
        """The maximum age of the cookie in seconds. Defaults to `None`."""
        return self.get("max-age")

    @max_age.setter
    def max_age(self, value: int) -> None:  # no cov
        self._set_value("max-age", value)

    @property
    def secure(self) -> bool:  # no cov
        """Whether the cookie is secure. Defaults to `True`."""
        return self.get("secure", False)

    @secure.setter
    def secure(self, value: bool) -> None:  # no cov
        self._set_value("secure", value)

    @property
    def httponly(self) -> bool:  # no cov
        """Whether the cookie is HTTP only. Defaults to `False`."""
        return self.get("httponly", False)

    @httponly.setter
    def httponly(self, value: bool) -> None:  # no cov
        self._set_value("httponly", value)

    @property
    def samesite(self) -> Optional[SameSite]:  # no cov
        """The SameSite attribute for the cookie. Defaults to `"Lax"`."""
        return self.get("samesite")

    @samesite.setter
    def samesite(self, value: SameSite) -> None:  # no cov
        self._set_value("samesite", value)

    @property
    def partitioned(self) -> bool:  # no cov
        """Whether the cookie is partitioned. Defaults to `False`."""
        return self.get("partitioned", False)

    @partitioned.setter
    def partitioned(self, value: bool) -> None:  # no cov
        self._set_value("partitioned", value)

    @classmethod
    def make_key(
        cls, key: str, host_prefix: bool = False, secure_prefix: bool = False
    ) -> str:
        """Create a cookie key with the appropriate prefix.

        Cookies can have one ow two prefixes. The first is `__Host-` which
        requires that the cookie be set with `path="/", domain=None, and
        secure=True`. The second is `__Secure-` which requires that
        `secure=True`.

        They cannot be combined.

        Args:
            key (str): The key (name) of the cookie.
            host_prefix (bool, optional): Whether to add __Host- as a prefix to the key.
                This requires that path="/", domain=None, and secure=True.
                Defaults to `False`.
            secure_prefix (bool, optional): Whether to add __Secure- as a prefix to the key.
                This requires that secure=True. Defaults to `False`.

        Raises:
            ServerError: If both host_prefix and secure_prefix are set.

        Returns:
            str: The key with the appropriate prefix.
        """  # noqa: E501
        if host_prefix and secure_prefix:
            raise ServerError(
                "Both host_prefix and secure_prefix were requested. "
                "A cookie should have only one prefix."
            )
        elif host_prefix:
            key = cls.HOST_PREFIX + key
        elif secure_prefix:
            key = cls.SECURE_PREFIX + key
        return key



from __future__ import annotations

from typing import Any, List, Optional


class RequestParameters(dict):
    """Hosts a dict with lists as values where get returns the first value of the list and getlist returns the whole shebang"""  # noqa: E501

    def get(self, name: str, default: Optional[Any] = None) -> Optional[Any]:
        """Return the first value, either the default or actual

        Args:
            name (str): The name of the parameter
            default (Optional[Any], optional): The default value. Defaults to None.

        Returns:
            Optional[Any]: The first value of the list
        """  # noqa: E501
        return super().get(name, [default])[0]

    def getlist(
        self, name: str, default: Optional[List[Any]] = None
    ) -> List[Any]:
        """Return the entire list

        Args:
            name (str): The name of the parameter
            default (Optional[List[Any]], optional): The default value. Defaults to None.

        Returns:
            list[Any]: The entire list of values or [] if not found
        """  # noqa: E501
        return super().get(name, default) or []



from sanic_testing.reusable import ReusableClient

from sanic.response import json, text


def test_no_body_requests(app, port):
    @app.get("/")
    async def handler(request):
        return json(
            {
                "request_id": str(request.id),
                "connection_id": id(request.conn_info),
            }
        )

    client = ReusableClient(app, port=port)

    with client:
        _, response1 = client.get("/")
        _, response2 = client.get("/")

    assert response1.status == response2.status == 200
    assert response1.json["request_id"] != response2.json["request_id"]
    assert response1.json["connection_id"] == response2.json["connection_id"]


def test_json_body_requests(app, port):
    @app.post("/")
    async def handler(request):
        return json(
            {
                "request_id": str(request.id),
                "connection_id": id(request.conn_info),
                "foo": request.json.get("foo"),
            }
        )

    client = ReusableClient(app, port=port)

    with client:
        _, response1 = client.post("/", json={"foo": True})
        _, response2 = client.post("/", json={"foo": True})

    assert response1.status == response2.status == 200
    assert response1.json["foo"] is response2.json["foo"] is True
    assert response1.json["request_id"] != response2.json["request_id"]
    assert response1.json["connection_id"] == response2.json["connection_id"]


def test_streaming_body_requests(app, port):
    @app.post("/", stream=True)
    async def handler(request):
        data = [part.decode("utf-8") async for part in request.stream]
        return json(
            {
                "request_id": str(request.id),
                "connection_id": id(request.conn_info),
                "data": data,
            }
        )

    data = ["hello", "world"]

    client = ReusableClient(app, port=port)

    async def stream(data):
        for value in data:
            yield value.encode("utf-8")

    with client:
        _, response1 = client.post("/", data=stream(data))
        _, response2 = client.post("/", data=stream(data))

    assert response1.status == response2.status == 200
    assert response1.json["data"] == response2.json["data"] == data
    assert response1.json["request_id"] != response2.json["request_id"]
    assert response1.json["connection_id"] == response2.json["connection_id"]


def test_bad_headers(app, port):
    @app.get("/")
    async def handler(request):
        return text("")

    @app.on_response
    async def reqid(request, response):
        response.headers["x-request-id"] = request.id

    client = ReusableClient(app, port=port)
    bad_headers = {"bad": "bad" * 5_000}

    with client:
        _, response1 = client.get("/")
        _, response2 = client.get("/", headers=bad_headers)

    assert response1.status == 200
    assert response2.status == 413
    assert (
        response1.headers["x-request-id"] != response2.headers["x-request-id"]
    )



import logging
import os

from pathlib import Path
from unittest.mock import Mock

import pytest

from sanic import Sanic
from sanic.application.state import ApplicationServerInfo


@pytest.fixture(autouse=True)
def no_skip():
    should_auto_reload = Sanic.should_auto_reload
    Sanic.should_auto_reload = Mock(return_value=False)
    yield
    Sanic._app_registry = {}
    Sanic.should_auto_reload = should_auto_reload
    try:
        del os.environ["SANIC_MOTD_OUTPUT"]
    except KeyError:
        ...


def get_primary(app: Sanic) -> ApplicationServerInfo:
    return app.state.server_info[0]


def test_dev(app: Sanic):
    app.prepare(dev=True)

    assert app.state.is_debug
    assert app.state.auto_reload


def test_motd_display(app: Sanic):
    app.prepare(motd_display={"foo": "bar"})

    assert app.config.MOTD_DISPLAY["foo"] == "bar"
    del app.config.MOTD_DISPLAY["foo"]


@pytest.mark.parametrize("dirs", ("./foo", ("./foo", "./bar")))
def test_reload_dir(app: Sanic, dirs, caplog):
    messages = []
    with caplog.at_level(logging.WARNING):
        app.prepare(reload_dir=dirs)

    if isinstance(dirs, str):
        dirs = (dirs,)
        for d in dirs:
            assert Path(d) in app.state.reload_dirs
            messages.append(
                f"Directory {d} could not be located",
            )

    for message in messages:
        assert ("sanic.root", logging.WARNING, message) in caplog.record_tuples


def test_fast(app: Sanic, caplog):
    @app.after_server_start
    async def stop(app, _):
        app.stop()

    try:
        workers = len(os.sched_getaffinity(0))
    except AttributeError:
        workers = os.cpu_count() or 1

    with caplog.at_level(logging.INFO):
        app.prepare(fast=True)

    assert app.state.fast
    assert app.state.workers == workers

    messages = [m[2] for m in caplog.record_tuples]

    if workers == 1:
        worker_fragment = "single worker"
    else:
        worker_fragment = f"w/ {workers} workers"

    assert f"mode: production, goin' fast {worker_fragment}" in messages



from importlib import import_module

import pytest


@pytest.mark.parametrize(
    "item",
    (
        "__version__",
        "Sanic",
        "Blueprint",
        "HTTPMethod",
        "HTTPResponse",
        "Request",
        "Websocket",
        "empty",
        "file",
        "html",
        "json",
        "redirect",
        "text",
    ),
)
def test_imports(item):
    import_module("sanic", item)



import asyncio

from asyncio.tasks import Task
from unittest.mock import Mock, call

import pytest

from sanic.app import Sanic
from sanic.application.state import ApplicationServerInfo, ServerStage
from sanic.response import empty


try:
    from unittest.mock import AsyncMock
except ImportError:
    from tests.asyncmock import AsyncMock  # type: ignore

pytestmark = pytest.mark.asyncio


async def dummy(n=0):
    for _ in range(n):
        await asyncio.sleep(1)
    return True


@pytest.fixture(autouse=True)
def mark_app_running(app: Sanic):
    app.state.server_info.append(
        ApplicationServerInfo(
            stage=ServerStage.SERVING, settings={}, server=AsyncMock()
        )
    )


async def test_add_task_returns_task(app: Sanic):
    task = app.add_task(dummy())

    assert isinstance(task, Task)
    assert len(app._task_registry) == 0


async def test_add_task_with_name(app: Sanic):
    task = app.add_task(dummy(), name="dummy")

    assert isinstance(task, Task)
    assert len(app._task_registry) == 1
    assert task is app.get_task("dummy")

    for task in app.tasks:
        assert task in app._task_registry.values()


async def test_cancel_task(app: Sanic):
    task = app.add_task(dummy(3), name="dummy")

    assert task
    assert not task.done()
    assert not task.cancelled()

    await asyncio.sleep(0.1)

    assert not task.done()
    assert not task.cancelled()

    await app.cancel_task("dummy")

    assert task.cancelled()


async def test_purge_tasks(app: Sanic):
    app.add_task(dummy(3), name="dummy")

    await app.cancel_task("dummy")

    assert len(app._task_registry) == 1

    app.purge_tasks()

    assert len(app._task_registry) == 0


async def test_purge_tasks_with_create_task(app: Sanic):
    app.add_task(asyncio.create_task(dummy(3)), name="dummy")

    await app.cancel_task("dummy")

    assert len(app._task_registry) == 1

    app.purge_tasks()

    assert len(app._task_registry) == 0


def test_shutdown_tasks_on_app_stop():
    class TestSanic(Sanic):
        shutdown_tasks = Mock()

    app = TestSanic("Test")

    @app.route("/")
    async def handler(_):
        return empty()

    app.test_client.get("/")

    app.shutdown_tasks.call_args == [
        call(timeout=0),
        call(15.0),
    ]



import logging
import os

from contextlib import contextmanager
from os import environ
from pathlib import Path
from tempfile import TemporaryDirectory
from textwrap import dedent
from unittest.mock import Mock, call

import pytest

from pytest import MonkeyPatch

from sanic import Sanic
from sanic.config import DEFAULT_CONFIG, Config
from sanic.constants import LocalCertCreator
from sanic.exceptions import PyFileError

from .conftest import get_port


@contextmanager
def temp_path():
    """a simple cross platform replacement for NamedTemporaryFile"""
    with TemporaryDirectory() as td:
        yield Path(td, "file")


class ConfigTest:
    not_for_config = "should not be used"
    CONFIG_VALUE = "should be used"

    @property
    def ANOTHER_VALUE(self):
        return self.CONFIG_VALUE

    @property
    def another_not_for_config(self):
        return self.not_for_config


class UltimateAnswer:
    def __init__(self, answer):
        self.answer = int(answer)


def test_load_from_object(app: Sanic):
    app.config.load(ConfigTest)
    assert "CONFIG_VALUE" in app.config
    assert app.config.CONFIG_VALUE == "should be used"
    assert "not_for_config" not in app.config


def test_load_from_object_string(app: Sanic):
    app.config.load("tests.test_config.ConfigTest")
    assert "CONFIG_VALUE" in app.config
    assert app.config.CONFIG_VALUE == "should be used"
    assert "not_for_config" not in app.config


def test_load_from_instance(app: Sanic):
    app.config.load(ConfigTest())
    assert "CONFIG_VALUE" in app.config
    assert app.config.CONFIG_VALUE == "should be used"
    assert app.config.ANOTHER_VALUE == "should be used"
    assert "not_for_config" not in app.config
    assert "another_not_for_config" not in app.config


def test_load_from_object_string_exception(app: Sanic):
    with pytest.raises(ImportError):
        app.config.load("test_config.Config.test")


def test_auto_env_prefix():
    environ["SANIC_TEST_ANSWER"] = "42"
    app = Sanic(name="Test")
    assert app.config.TEST_ANSWER == 42
    del environ["SANIC_TEST_ANSWER"]


def test_auto_bool_env_prefix():
    environ["SANIC_TEST_ANSWER"] = "True"
    app = Sanic(name="Test")
    assert app.config.TEST_ANSWER is True
    del environ["SANIC_TEST_ANSWER"]


@pytest.mark.parametrize("env_prefix", [None, ""])
def test_empty_load_env_prefix(env_prefix):
    environ["SANIC_TEST_ANSWER"] = "42"
    app = Sanic(name="Test", env_prefix=env_prefix)
    assert getattr(app.config, "TEST_ANSWER", None) is None
    del environ["SANIC_TEST_ANSWER"]


def test_env_prefix():
    environ["MYAPP_TEST_ANSWER"] = "42"
    app = Sanic(name="Test", env_prefix="MYAPP_")
    assert app.config.TEST_ANSWER == 42
    del environ["MYAPP_TEST_ANSWER"]


def test_env_prefix_float_values():
    environ["MYAPP_TEST_ROI"] = "2.3"
    app = Sanic(name="Test", env_prefix="MYAPP_")
    assert app.config.TEST_ROI == 2.3
    del environ["MYAPP_TEST_ROI"]


def test_env_prefix_string_value():
    environ["MYAPP_TEST_TOKEN"] = "somerandomtesttoken"
    app = Sanic(name="Test", env_prefix="MYAPP_")
    assert app.config.TEST_TOKEN == "somerandomtesttoken"
    del environ["MYAPP_TEST_TOKEN"]


def test_env_w_custom_converter():
    environ["SANIC_TEST_ANSWER"] = "42"

    config = Config(converters=[UltimateAnswer])
    app = Sanic(name="Test", config=config)
    assert isinstance(app.config.TEST_ANSWER, UltimateAnswer)
    assert app.config.TEST_ANSWER.answer == 42
    del environ["SANIC_TEST_ANSWER"]


def test_env_lowercase():
    environ["SANIC_test_answer"] = "42"
    app = Sanic(name="Test")
    assert "test_answer" not in app.config
    del environ["SANIC_test_answer"]


def test_add_converter_multiple_times(caplog):
    def converter(): ...

    message = (
        "Configuration value converter 'converter' has already been registered"
    )
    config = Config()
    config.register_type(converter)
    with caplog.at_level(logging.WARNING):
        config.register_type(converter)

    assert ("sanic.error", logging.WARNING, message) in caplog.record_tuples
    assert len(config._converters) == 5


def test_load_from_file(app: Sanic):
    config = dedent(
        """
    VALUE = 'some value'
    condition = 1 == 1
    if condition:
        CONDITIONAL = 'should be set'
    """
    )
    with temp_path() as config_path:
        config_path.write_text(config)
        app.config.load(str(config_path))
        assert "VALUE" in app.config
        assert app.config.VALUE == "some value"
        assert "CONDITIONAL" in app.config
        assert app.config.CONDITIONAL == "should be set"
        assert "condition" not in app.config


def test_load_from_missing_file(app: Sanic):
    with pytest.raises(IOError):
        app.config.load("non-existent file")


def test_load_from_envvar(app: Sanic):
    config = "VALUE = 'some value'"
    with temp_path() as config_path:
        config_path.write_text(config)
        environ["APP_CONFIG"] = str(config_path)
        app.config.load("${APP_CONFIG}")
        assert "VALUE" in app.config
        assert app.config.VALUE == "some value"


def test_load_from_missing_envvar(app: Sanic):
    with pytest.raises(IOError) as e:
        app.config.load("non-existent variable")
        assert str(e.value) == (
            "The environment variable 'non-existent "
            "variable' is not set and thus configuration "
            "could not be loaded."
        )


def test_load_config_from_file_invalid_syntax(app: Sanic):
    config = "VALUE = some value"
    with temp_path() as config_path:
        config_path.write_text(config)

        with pytest.raises(PyFileError):
            app.config.load(config_path)


def test_overwrite_exisiting_config(app: Sanic):
    app.config.DEFAULT = 1

    class Config:
        DEFAULT = 2

    app.config.load(Config)
    assert app.config.DEFAULT == 2


def test_overwrite_exisiting_config_ignore_lowercase(app: Sanic):
    app.config.default = 1

    class Config:
        default = 2

    app.config.load(Config)
    assert app.config.default == 1


def test_missing_config(app: Sanic):
    with pytest.raises(AttributeError, match="Config has no 'NON_EXISTENT'"):
        _ = app.config.NON_EXISTENT


def test_config_defaults():
    """
    load DEFAULT_CONFIG
    """
    conf = Config()
    for key, value in DEFAULT_CONFIG.items():
        assert getattr(conf, key) == value


def test_config_custom_defaults():
    """
    we should have all the variables from defaults rewriting them with
    custom defaults passed in
    Config
    """
    custom_defaults = {
        "REQUEST_MAX_SIZE": 1,
        "KEEP_ALIVE": False,
        "ACCESS_LOG": False,
    }
    conf = Config(defaults=custom_defaults)
    for key, value in DEFAULT_CONFIG.items():
        if key in custom_defaults.keys():
            value = custom_defaults[key]
        assert getattr(conf, key) == value


def test_config_custom_defaults_with_env():
    """
    test that environment variables has higher priority than DEFAULT_CONFIG
    and passed defaults dict
    """
    custom_defaults = {
        "REQUEST_MAX_SIZE123": 1,
        "KEEP_ALIVE123": False,
        "ACCESS_LOG123": False,
    }

    environ_defaults = {
        "SANIC_REQUEST_MAX_SIZE123": "2",
        "SANIC_KEEP_ALIVE123": "True",
        "SANIC_ACCESS_LOG123": "False",
    }

    for key, value in environ_defaults.items():
        environ[key] = value

    conf = Config(defaults=custom_defaults)
    for key, value in DEFAULT_CONFIG.items():
        if "SANIC_" + key in environ_defaults.keys():
            value = environ_defaults["SANIC_" + key]
            try:
                value = int(value)
            except ValueError:
                if value in ["True", "False"]:
                    value = value == "True"

        assert getattr(conf, key) == value

    for key, value in environ_defaults.items():
        del environ[key]


@pytest.mark.parametrize("access_log", (True, False))
def test_config_access_log_passing_in_run(app: Sanic, access_log):
    assert app.config.ACCESS_LOG is False

    @app.listener("after_server_start")
    async def _request(sanic, loop):
        app.stop()

    app.run(port=1340, access_log=access_log, single_process=True)
    assert app.config.ACCESS_LOG is access_log


@pytest.mark.asyncio
async def test_config_access_log_passing_in_create_server(app: Sanic):
    assert app.config.ACCESS_LOG is False

    @app.listener("after_server_start")
    async def _request(sanic, loop):
        app.stop()

    await app.create_server(
        port=get_port(), access_log=False, return_asyncio_server=True
    )
    assert app.config.ACCESS_LOG is False

    await app.create_server(
        port=get_port(), access_log=True, return_asyncio_server=True
    )
    assert app.config.ACCESS_LOG is True


def test_config_rewrite_keep_alive():
    config = Config()
    assert config.KEEP_ALIVE == DEFAULT_CONFIG["KEEP_ALIVE"]
    config = Config(keep_alive=True)
    assert config.KEEP_ALIVE is True
    config = Config(keep_alive=False)
    assert config.KEEP_ALIVE is False

    # use defaults
    config = Config(defaults={"KEEP_ALIVE": False})
    assert config.KEEP_ALIVE is False
    config = Config(defaults={"KEEP_ALIVE": True})
    assert config.KEEP_ALIVE is True


_test_setting_as_dict = {"TEST_SETTING_VALUE": 1}
_test_setting_as_class = type("C", (), {"TEST_SETTING_VALUE": 1})
_test_setting_as_module = str(
    Path(__file__).parent / "static/app_test_config.py"
)


@pytest.mark.parametrize(
    "conf_object",
    [
        _test_setting_as_dict,
        _test_setting_as_class,
        _test_setting_as_module,
    ],
    ids=["from_dict", "from_class", "from_file"],
)
def test_update(app: Sanic, conf_object):
    app.update_config(conf_object)
    assert app.config["TEST_SETTING_VALUE"] == 1


def test_update_from_lowercase_key(app: Sanic):
    d = {"test_setting_value": 1}
    app.update_config(d)
    assert "test_setting_value" not in app.config


def test_config_set_methods(app: Sanic, monkeypatch: MonkeyPatch):
    post_set = Mock()
    monkeypatch.setattr(Config, "_post_set", post_set)

    app.config.FOO = 1
    post_set.assert_called_once_with("FOO", 1)
    post_set.reset_mock()

    app.config["FOO"] = 2
    post_set.assert_called_once_with("FOO", 2)
    post_set.reset_mock()

    app.config.update({"FOO": 3})
    post_set.assert_called_once_with("FOO", 3)
    post_set.reset_mock()

    app.config.update([("FOO", 4)])
    post_set.assert_called_once_with("FOO", 4)
    post_set.reset_mock()

    app.config.update(FOO=5)
    post_set.assert_called_once_with("FOO", 5)
    post_set.reset_mock()

    app.config.update({"FOO": 6}, {"BAR": 7})
    post_set.assert_has_calls(
        calls=[
            call("FOO", 6),
            call("BAR", 7),
        ]
    )
    post_set.reset_mock()

    app.config.update({"FOO": 8}, BAR=9)
    post_set.assert_has_calls(
        calls=[
            call("FOO", 8),
            call("BAR", 9),
        ],
        any_order=True,
    )
    post_set.reset_mock()

    app.config.update_config({"FOO": 10})
    post_set.assert_called_once_with("FOO", 10)


def test_negative_proxy_count(app: Sanic):
    app.config.PROXIES_COUNT = -1

    message = (
        "PROXIES_COUNT cannot be negative. "
        "https://sanic.readthedocs.io/en/latest/sanic/config.html"
        "#proxy-configuration"
    )
    with pytest.raises(ValueError, match=message):
        app.prepare()


@pytest.mark.parametrize(
    "passed,expected",
    (
        ("auto", LocalCertCreator.AUTO),
        ("mkcert", LocalCertCreator.MKCERT),
        ("trustme", LocalCertCreator.TRUSTME),
        ("AUTO", LocalCertCreator.AUTO),
        ("MKCERT", LocalCertCreator.MKCERT),
        ("TRUSTME", LocalCertCreator.TRUSTME),
    ),
)
def test_convert_local_cert_creator(passed, expected):
    os.environ["SANIC_LOCAL_CERT_CREATOR"] = passed
    app = Sanic("Test")
    assert app.config.LOCAL_CERT_CREATOR is expected
    del os.environ["SANIC_LOCAL_CERT_CREATOR"]



import sys

from multiprocessing import set_start_method
from unittest.mock import patch

import pytest

from sanic import Sanic


@pytest.mark.parametrize(
    "start_method,platform,expected",
    (
        (None, "linux", "spawn"),
        (None, "other", "spawn"),
        ("fork", "linux", "fork"),
        ("fork", "other", "fork"),
        ("forkserver", "linux", "forkserver"),
        ("forkserver", "other", "forkserver"),
        ("spawn", "linux", "spawn"),
        ("spawn", "other", "spawn"),
    ),
)
def test_get_context(start_method, platform, expected):
    if start_method:
        Sanic.start_method = start_method
    with patch("sys.platform", platform):
        assert Sanic._get_startup_method() == expected


@pytest.mark.skipif(
    not sys.platform.startswith("linux"), reason="Only test on Linux"
)
def test_set_startup_catch():
    Sanic.START_METHOD_SET = False
    set_start_method("fork", force=True)
    Sanic.test_mode = False
    message = (
        "Start method 'spawn' was requested, but 'fork' was already set.\n"
        "For more information, see: https://sanic.dev/en/guide/running/manager.html#overcoming-a-coderuntimeerrorcode"
    )
    with pytest.raises(RuntimeError, match=message):
        Sanic._set_startup_method()
    Sanic.test_mode = True



import pytest

from sanic import Sanic
from sanic.log import deprecation


def test_deprecation():
    message = r"\[DEPRECATION v9\.9\] hello"
    with pytest.warns(DeprecationWarning, match=message):
        deprecation("hello", 9.9)


@pytest.mark.parametrize(
    "filter,expected",
    (("default", 1), ("once", 1), ("ignore", 0)),
)
def test_deprecation_filter(app: Sanic, filter, expected, recwarn):
    app.config.DEPRECATION_FILTER = filter
    deprecation("hello", 9.9)
    assert len(recwarn) == expected



from unittest.mock import Mock

import pytest

from sanic.worker.constants import ProcessState, RestartOrder
from sanic.worker.process import WorkerProcess
from sanic.worker.restarter import Restarter


def noop(*args, **kwargs):
    pass


def make_worker_process(
    name: str, state: ProcessState = ProcessState.STARTED
) -> WorkerProcess:
    worker_process = Mock()
    worker_process.restart = Mock()
    worker_process.name = name
    worker_process.state = state
    return worker_process


def test_restart_transient():
    transient = make_worker_process("Transient")
    durable = make_worker_process("Durable")
    restarter = Restarter()

    restarter.restart([transient], [durable])
    transient.restart.assert_called_once_with(
        restart_order=RestartOrder.SHUTDOWN_FIRST
    )
    durable.restart.assert_not_called()
    transient.restart.reset_mock()
    restarter.restart(
        [transient], [durable], restart_order=RestartOrder.STARTUP_FIRST
    )
    transient.restart.assert_called_once_with(
        restart_order=RestartOrder.STARTUP_FIRST
    )


@pytest.mark.parametrize(
    "state,called",
    (
        (ProcessState.IDLE, False),
        (ProcessState.RESTARTING, False),
        (ProcessState.STARTING, False),
        (ProcessState.STARTED, False),
        (ProcessState.ACKED, False),
        (ProcessState.JOINED, False),
        (ProcessState.TERMINATED, False),
        (ProcessState.FAILED, True),
        (ProcessState.COMPLETED, True),
        (ProcessState.NONE, True),
    ),
)
def test_restart_durable(caplog, state, called):
    transient = make_worker_process("Transient")
    durable = make_worker_process("Durable")
    restarter = Restarter()

    restarter.restart([transient], [durable], process_names=["Durable"])

    transient.restart.assert_not_called()
    durable.restart.assert_not_called()

    assert (
        "sanic.error",
        40,
        "Cannot restart process Durable because it is not in a "
        "final state. Current state is: STARTED.",
    ) in caplog.record_tuples
    assert (
        "sanic.error",
        40,
        "Failed to restart processes: Durable",
    ) in caplog.record_tuples

    durable.state = state
    restarter.restart([transient], [durable], process_names=["Durable"])

    transient.restart.assert_not_called()
    if called:
        durable.restart.assert_called_once_with(
            restart_order=RestartOrder.SHUTDOWN_FIRST
        )
    else:
        durable.restart.assert_not_called()



from ast import Attribute, Await, Expr, NodeTransformer
from typing import Any, List

from sanic.log import logger

from .base import BaseScheme


class OptionalDispatchEvent(BaseScheme):
    ident = "ODE"
    SYNC_SIGNAL_NAMESPACES = "http."

    def __init__(self, app) -> None:
        super().__init__(app)

        self._sync_events()
        self._registered_events = [
            signal.name for signal in app.signal_router.routes
        ]

    def visitors(self) -> List[NodeTransformer]:
        return [RemoveDispatch(self._registered_events)]

    def _sync_events(self):
        all_events = set()
        app_events = {}
        for app in self.app.__class__._app_registry.values():
            if app.state.server_info:
                app_events[app] = {
                    signal.name for signal in app.signal_router.routes
                }
                all_events.update(app_events[app])

        for app, events in app_events.items():
            missing = {
                x
                for x in all_events.difference(events)
                if any(x.startswith(y) for y in self.SYNC_SIGNAL_NAMESPACES)
            }
            if missing:
                was_finalized = app.signal_router.finalized
                if was_finalized:  # no cov
                    app.signal_router.reset()
                for event in missing:
                    app.signal(event)(self.noop)
                if was_finalized:  # no cov
                    app.signal_router.finalize()

    @staticmethod
    async def noop(**_):  # no cov
        ...


class RemoveDispatch(NodeTransformer):
    def __init__(self, registered_events) -> None:
        self._registered_events = registered_events

    def visit_Expr(self, node: Expr) -> Any:
        call = node.value
        if isinstance(call, Await):
            call = call.value

        func = getattr(call, "func", None)
        args = getattr(call, "args", None)
        if not func or not args:
            return node

        if isinstance(func, Attribute) and func.attr == "dispatch":
            event = args[0]
            if event_name := getattr(event, "value", None):
                if self._not_registered(event_name):
                    logger.debug(
                        f"Disabling event: {event_name}",
                        extra={"verbosity": 2},
                    )
                    return None
        return node

    def _not_registered(self, event_name):
        dynamic = []
        for event in self._registered_events:
            if event.endswith(">"):
                namespace_concern, _ = event.rsplit(".", 1)
                dynamic.append(namespace_concern)

        namespace_concern, _ = event_name.rsplit(".", 1)
        return (
            event_name not in self._registered_events
            and namespace_concern not in dynamic
        )



import os

from types import SimpleNamespace
from typing import Any, Iterable

from sanic.log import Colors, error_logger


class SharedContext(SimpleNamespace):
    SAFE = ("_lock",)

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._lock = False

    def __setattr__(self, name: str, value: Any) -> None:
        if self.is_locked:
            raise RuntimeError(
                f"Cannot set {name} on locked SharedContext object"
            )
        if not os.environ.get("SANIC_WORKER_NAME"):
            to_check: Iterable[Any]
            if not isinstance(value, (tuple, frozenset)):
                to_check = [value]
            else:
                to_check = value
            for item in to_check:
                self._check(name, item)
        super().__setattr__(name, value)

    def _check(self, name: str, value: Any) -> None:
        if name in self.SAFE:
            return
        try:
            module = value.__module__
        except AttributeError:
            module = ""
        if not any(
            module.startswith(prefix)
            for prefix in ("multiprocessing", "ctypes")
        ):
            error_logger.warning(
                f"{Colors.YELLOW}Unsafe object {Colors.PURPLE}{name} "
                f"{Colors.YELLOW}with type {Colors.PURPLE}{type(value)} "
                f"{Colors.YELLOW}was added to shared_ctx. It may not "
                "not function as intended. Consider using the regular "
                f"ctx.\nFor more information, please see https://sanic.dev/en"
                "/guide/deployment/manager.html#using-shared-context-between-"
                f"worker-processes.{Colors.END}"
            )

    @property
    def is_locked(self) -> bool:
        return getattr(self, "_lock", False)

    def lock(self) -> None:
        self._lock = True



import logging

from unittest.mock import Mock, patch

import pytest

from sanic.compat import OS_IS_WINDOWS, UVLOOP_INSTALLED
from sanic.server import loop


@pytest.mark.skipif(
    not OS_IS_WINDOWS,
    reason="Not testable with current client",
)
def test_raises_warning_if_os_is_windows(caplog):
    with caplog.at_level(logging.WARNING):
        loop.try_use_uvloop()

    for record in caplog.records:
        if record.message.startswith("You are trying to use"):
            break

    assert record.message == (
        "You are trying to use uvloop, but uvloop is not compatible "
        "with your system. You can disable uvloop completely by setting "
        "the 'USE_UVLOOP' configuration value to false, or simply not "
        "defining it and letting Sanic handle it for you. Sanic will now "
        "continue to run using the default event loop."
    )


@pytest.mark.skipif(
    OS_IS_WINDOWS or UVLOOP_INSTALLED,
    reason="Not testable with current client",
)
def test_raises_warning_if_uvloop_not_installed(caplog):
    with caplog.at_level(logging.WARNING):
        loop.try_use_uvloop()

    for record in caplog.records:
        if record.message.startswith("You are trying to use"):
            break

    assert record.message == (
        "You are trying to use uvloop, but uvloop is not "
        "installed in your system. In order to use uvloop "
        "you must first install it. Otherwise, you can disable "
        "uvloop completely by setting the 'USE_UVLOOP' "
        "configuration value to false. Sanic will now continue "
        "to run with the default event loop."
    )


@pytest.mark.skipif(
    OS_IS_WINDOWS or not UVLOOP_INSTALLED,
    reason="Not testable with current client",
)
def test_logs_when_install_and_runtime_config_mismatch(caplog, monkeypatch):
    getenv = Mock(return_value="no")
    monkeypatch.setattr(loop, "getenv", getenv)

    with caplog.at_level(logging.INFO):
        loop.try_use_uvloop()

    getenv.assert_called_once_with("SANIC_NO_UVLOOP", "no")
    assert caplog.record_tuples == []

    getenv = Mock(return_value="yes")
    monkeypatch.setattr(loop, "getenv", getenv)
    with caplog.at_level(logging.INFO):
        loop.try_use_uvloop()

    getenv.assert_called_once_with("SANIC_NO_UVLOOP", "no")
    for record in caplog.records:
        if record.message.startswith("You are requesting to run"):
            break

    assert record.message == (
        "You are requesting to run Sanic using uvloop, but the "
        "install-time 'SANIC_NO_UVLOOP' environment variable (used to "
        "opt-out of installing uvloop with Sanic) is set to true. If "
        "you want to prevent Sanic from overriding the event loop policy "
        "during runtime, set the 'USE_UVLOOP' configuration value to "
        "false."
    )


@pytest.mark.skipif(
    OS_IS_WINDOWS or not UVLOOP_INSTALLED,
    reason="Not testable with current client",
)
def test_sets_loop_policy_only_when_not_already_set(monkeypatch):
    import uvloop  # type: ignore

    # Existing policy is not uvloop.EventLoopPolicy
    get_event_loop_policy = Mock(return_value=None)
    monkeypatch.setattr(
        loop.asyncio, "get_event_loop_policy", get_event_loop_policy
    )

    with patch("asyncio.set_event_loop_policy") as set_event_loop_policy:
        loop.try_use_uvloop()
        set_event_loop_policy.assert_called_once()
        args, _ = set_event_loop_policy.call_args
        policy = args[0]
        assert isinstance(policy, uvloop.EventLoopPolicy)

    # Existing policy is uvloop.EventLoopPolicy
    get_event_loop_policy = Mock(return_value=policy)
    monkeypatch.setattr(
        loop.asyncio, "get_event_loop_policy", get_event_loop_policy
    )

    with patch("asyncio.set_event_loop_policy") as set_event_loop_policy:
        loop.try_use_uvloop()
        set_event_loop_policy.assert_not_called()



from os import environ
from pathlib import Path
from types import ModuleType

import pytest

from sanic.exceptions import LoadFileException
from sanic.utils import load_module_from_file_location


@pytest.mark.parametrize(
    "location",
    (
        Path(__file__).parent / "static" / "app_test_config.py",
        str(Path(__file__).parent / "static" / "app_test_config.py"),
        str(Path(__file__).parent / "static" / "app_test_config.py").encode(),
    ),
)
def test_load_module_from_file_location(location):
    module = load_module_from_file_location(location)

    assert isinstance(module, ModuleType)


def test_loaded_module_from_file_location_name():
    module = load_module_from_file_location(
        str(Path(__file__).parent / "static" / "app_test_config.py")
    )

    name = module.__name__
    if "C:\\" in name:
        name = name.split("\\")[-1]
    assert name == "app_test_config"


def test_load_module_from_file_location_with_non_existing_env_variable():
    with pytest.raises(
        LoadFileException,
        match="The following environment variables are not set: MuuMilk",
    ):
        load_module_from_file_location("${MuuMilk}")


def test_load_module_from_file_location_using_env():
    environ["APP_TEST_CONFIG"] = "static/app_test_config.py"
    location = str(Path(__file__).parent / "${APP_TEST_CONFIG}")
    module = load_module_from_file_location(location)

    assert isinstance(module, ModuleType)



# import logging

# from unittest.mock import Mock

# import pytest

# from sanic import Sanic
# from sanic.response import text
# from sanic.server.async_server import AsyncioServer
# from sanic.signals import Event
# from sanic.touchup.schemes.ode import OptionalDispatchEvent


# try:
#     from unittest.mock import AsyncMock
# except ImportError:
#     from tests.asyncmock import AsyncMock  # type: ignore


# @pytest.fixture
# def app_one():
#     app = Sanic("One")

#     @app.get("/one")
#     async def one(request):
#         return text("one")

#     return app


# @pytest.fixture
# def app_two():
#     app = Sanic("Two")

#     @app.get("/two")
#     async def two(request):
#         return text("two")

#     return app


# @pytest.fixture(autouse=True)
# def clean():
#     Sanic._app_registry = {}
#     yield


# def test_serve_same_app_multiple_tuples(app_one, run_multi):
#     app_one.prepare(port=23456)
#     app_one.prepare(port=23457)

#     logs = run_multi(app_one)
#     assert (
#         "sanic.root",
#         logging.INFO,
#         "Goin' Fast @ http://127.0.0.1:23456",
#     ) in logs
#     assert (
#         "sanic.root",
#         logging.INFO,
#         "Goin' Fast @ http://127.0.0.1:23457",
#     ) in logs


# def test_serve_multiple_apps(app_one, app_two, run_multi):
#     app_one.prepare(port=23456)
#     app_two.prepare(port=23457)

#     logs = run_multi(app_one)
#     assert (
#         "sanic.root",
#         logging.INFO,
#         "Goin' Fast @ http://127.0.0.1:23456",
#     ) in logs
#     assert (
#         "sanic.root",
#         logging.INFO,
#         "Goin' Fast @ http://127.0.0.1:23457",
#     ) in logs


# def test_listeners_on_secondary_app(app_one, app_two, run_multi):
#     app_one.prepare(port=23456)
#     app_two.prepare(port=23457)

#     before_start = AsyncMock()
#     after_start = AsyncMock()
#     before_stop = AsyncMock()
#     after_stop = AsyncMock()

#     app_two.before_server_start(before_start)
#     app_two.after_server_start(after_start)
#     app_two.before_server_stop(before_stop)
#     app_two.after_server_stop(after_stop)

#     run_multi(app_one)

#     before_start.assert_awaited_once()
#     after_start.assert_awaited_once()
#     before_stop.assert_awaited_once()
#     after_stop.assert_awaited_once()


# @pytest.mark.parametrize(
#     "events",
#     (
#         (Event.HTTP_LIFECYCLE_BEGIN,),
#         (Event.HTTP_LIFECYCLE_BEGIN, Event.HTTP_LIFECYCLE_COMPLETE),
#         (
#             Event.HTTP_LIFECYCLE_BEGIN,
#             Event.HTTP_LIFECYCLE_COMPLETE,
#             Event.HTTP_LIFECYCLE_REQUEST,
#         ),
#     ),
# )
# def test_signal_synchronization(app_one, app_two, run_multi, events):
#     app_one.prepare(port=23456)
#     app_two.prepare(port=23457)

#     for event in events:
#         app_one.signal(event)(AsyncMock())

#     run_multi(app_one)

#     assert len(app_two.signal_router.routes) == len(events) + 1

#     signal_handlers = {
#         signal.handler
#         for signal in app_two.signal_router.routes
#         if signal.name.startswith("http")
#     }

#     assert len(signal_handlers) == 1
#     assert list(signal_handlers)[0] is OptionalDispatchEvent.noop


# def test_warning_main_process_listeners_on_secondary(
#     app_one, app_two, run_multi
# ):
#     app_two.main_process_start(AsyncMock())
#     app_two.main_process_stop(AsyncMock())
#     app_one.prepare(port=23456)
#     app_two.prepare(port=23457)

#     log = run_multi(app_one)

#     message = (
#         f"Sanic found 2 listener(s) on "
#         "secondary applications attached to the main "
#         "process. These will be ignored since main "
#         "process listeners can only be attached to your "
#         "primary application: "
#         f"{repr(app_one)}"
#     )

#     assert ("sanic.error", logging.WARNING, message) in log


# def test_no_applications():
#     Sanic._app_registry = {}
#     message = "Did not find any applications."
#     with pytest.raises(RuntimeError, match=message):
#         Sanic.serve()


# def test_oserror_warning(app_one, app_two, run_multi, capfd):
#     orig = AsyncioServer.__await__
#     AsyncioServer.__await__ = Mock(side_effect=OSError("foo"))
#     app_one.prepare(port=23456, workers=2)
#     app_two.prepare(port=23457, workers=2)

#     run_multi(app_one)

#     captured = capfd.readouterr()
#     assert (
#         "An OSError was detected on startup. The encountered error was: foo"
#     ) in captured.err

#     AsyncioServer.__await__ = orig


# def test_running_multiple_offset_warning(app_one, app_two, run_multi, capfd):
#     app_one.prepare(port=23456, workers=2)
#     app_two.prepare(port=23457)

#     run_multi(app_one)

#     captured = capfd.readouterr()
#     assert (
#         f"The primary application {repr(app_one)} is running "
#         "with 2 worker(s). All "
#         "application instances will run with the same number. "
#         f"You requested {repr(app_two)} to run with "
#         "1 worker(s), which will be ignored "
#         "in favor of the primary application."
#     ) in captured.err


# def test_running_multiple_secondary(app_one, app_two, run_multi, capfd):
#     app_one.prepare(port=23456, workers=2)
#     app_two.prepare(port=23457)

#     before_start = AsyncMock()
#     app_two.before_server_start(before_start)
#     run_multi(app_one)

#     before_start.await_count == 2



import asyncio
import logging

from collections import deque, namedtuple
from unittest.mock import call

import pytest
import uvicorn

from httpx import Headers
from pytest import MonkeyPatch

from sanic import Sanic
from sanic.application.state import Mode
from sanic.asgi import Lifespan, MockTransport
from sanic.exceptions import BadRequest, Forbidden, ServiceUnavailable
from sanic.request import Request
from sanic.response import json, text
from sanic.server.websockets.connection import WebSocketConnection
from sanic.signals import RESERVED_NAMESPACES

from .conftest import get_port


try:
    from unittest.mock import AsyncMock
except ImportError:
    from tests.asyncmock import AsyncMock  # type: ignore


@pytest.fixture
def message_stack():
    return deque()


@pytest.fixture
def receive(message_stack):
    async def _receive():
        return message_stack.popleft()

    return _receive


@pytest.fixture
def send(message_stack):
    async def _send(message):
        message_stack.append(message)

    return _send


@pytest.fixture
def transport(message_stack, receive, send):
    return MockTransport({}, receive, send)


@pytest.fixture
def protocol(transport):
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    transport.loop = loop
    return transport.get_protocol()


def test_listeners_triggered(caplog):
    app = Sanic("app")
    before_server_start = False
    after_server_start = False
    before_server_stop = False
    after_server_stop = False

    @app.listener("before_server_start")
    def do_before_server_start(*args, **kwargs):
        nonlocal before_server_start
        before_server_start = True

    @app.listener("after_server_start")
    def do_after_server_start(*args, **kwargs):
        nonlocal after_server_start
        after_server_start = True

    @app.listener("before_server_stop")
    def do_before_server_stop(*args, **kwargs):
        nonlocal before_server_stop
        before_server_stop = True

    @app.listener("after_server_stop")
    def do_after_server_stop(*args, **kwargs):
        nonlocal after_server_stop
        after_server_stop = True

    @app.route("/")
    def handler(request):
        return text("...")

    class CustomServer(uvicorn.Server):
        def install_signal_handlers(self):
            pass

    config = uvicorn.Config(
        app=app, loop="asyncio", limit_max_requests=0, port=get_port()
    )
    server = CustomServer(config=config)

    start_message = (
        'You have set a listener for "before_server_start" in ASGI mode. '
        "It will be executed as early as possible, but not before the ASGI "
        "server is started."
    )
    stop_message = (
        'You have set a listener for "after_server_stop" in ASGI mode. '
        "It will be executed as late as possible, but not after the ASGI "
        "server is stopped."
    )

    with caplog.at_level(logging.DEBUG):
        server.run()

    assert (
        "sanic.root",
        logging.DEBUG,
        start_message,
    ) not in caplog.record_tuples
    assert (
        "sanic.root",
        logging.DEBUG,
        stop_message,
    ) not in caplog.record_tuples

    assert before_server_start
    assert after_server_start
    assert before_server_stop
    assert after_server_stop

    app.state.mode = Mode.DEBUG
    with caplog.at_level(logging.DEBUG):
        server.run()

    assert (
        "sanic.root",
        logging.DEBUG,
        start_message,
    ) not in caplog.record_tuples
    assert (
        "sanic.root",
        logging.DEBUG,
        stop_message,
    ) not in caplog.record_tuples

    app.state.verbosity = 2
    with caplog.at_level(logging.DEBUG):
        server.run()

    assert (
        "sanic.root",
        logging.DEBUG,
        start_message,
    ) in caplog.record_tuples
    assert (
        "sanic.root",
        logging.DEBUG,
        stop_message,
    ) in caplog.record_tuples


def test_listeners_triggered_async(app, caplog):
    before_server_start = False
    after_server_start = False
    before_server_stop = False
    after_server_stop = False

    @app.listener("before_server_start")
    async def do_before_server_start(*args, **kwargs):
        nonlocal before_server_start
        before_server_start = True

    @app.listener("after_server_start")
    async def do_after_server_start(*args, **kwargs):
        nonlocal after_server_start
        after_server_start = True

    @app.listener("before_server_stop")
    async def do_before_server_stop(*args, **kwargs):
        nonlocal before_server_stop
        before_server_stop = True

    @app.listener("after_server_stop")
    async def do_after_server_stop(*args, **kwargs):
        nonlocal after_server_stop
        after_server_stop = True

    @app.route("/")
    def handler(request):
        return text("...")

    class CustomServer(uvicorn.Server):
        def install_signal_handlers(self):
            pass

    config = uvicorn.Config(
        app=app, loop="asyncio", limit_max_requests=0, port=get_port()
    )
    server = CustomServer(config=config)

    start_message = (
        'You have set a listener for "before_server_start" in ASGI mode. '
        "It will be executed as early as possible, but not before the ASGI "
        "server is started."
    )
    stop_message = (
        'You have set a listener for "after_server_stop" in ASGI mode. '
        "It will be executed as late as possible, but not after the ASGI "
        "server is stopped."
    )

    with caplog.at_level(logging.DEBUG):
        server.run()

    assert (
        "sanic.root",
        logging.DEBUG,
        start_message,
    ) not in caplog.record_tuples
    assert (
        "sanic.root",
        logging.DEBUG,
        stop_message,
    ) not in caplog.record_tuples

    assert before_server_start
    assert after_server_start
    assert before_server_stop
    assert after_server_stop

    app.state.mode = Mode.DEBUG
    app.state.verbosity = 0
    with caplog.at_level(logging.DEBUG):
        server.run()

    assert (
        "sanic.root",
        logging.DEBUG,
        start_message,
    ) not in caplog.record_tuples
    assert (
        "sanic.root",
        logging.DEBUG,
        stop_message,
    ) not in caplog.record_tuples

    app.state.verbosity = 2
    with caplog.at_level(logging.DEBUG):
        server.run()

    assert (
        "sanic.root",
        logging.DEBUG,
        start_message,
    ) in caplog.record_tuples
    assert (
        "sanic.root",
        logging.DEBUG,
        stop_message,
    ) in caplog.record_tuples


def test_non_default_uvloop_config_raises_warning(app):
    app.config.USE_UVLOOP = True

    class CustomServer(uvicorn.Server):
        def install_signal_handlers(self):
            pass

    config = uvicorn.Config(
        app=app, loop="asyncio", limit_max_requests=0, port=get_port()
    )
    server = CustomServer(config=config)

    with pytest.warns(UserWarning) as records:
        server.run()

    msg = ""
    for record in records:
        _msg = str(record.message)
        if _msg.startswith("You have set the USE_UVLOOP configuration"):
            msg = _msg
            break

    assert msg == (
        "You have set the USE_UVLOOP configuration option, but Sanic "
        "cannot control the event loop when running in ASGI mode."
        "This option will be ignored."
    )


@pytest.mark.asyncio
async def test_mockprotocol_events(protocol):
    assert protocol._not_paused.is_set()
    protocol.pause_writing()
    assert not protocol._not_paused.is_set()
    protocol.resume_writing()
    assert protocol._not_paused.is_set()


@pytest.mark.asyncio
async def test_protocol_push_data(protocol, message_stack):
    text = b"hello"

    await protocol.push_data(text)
    await protocol.complete()

    assert len(message_stack) == 2

    message = message_stack.popleft()
    assert message["type"] == "http.response.body"
    assert message["more_body"]
    assert message["body"] == text

    message = message_stack.popleft()
    assert message["type"] == "http.response.body"
    assert not message["more_body"]
    assert message["body"] == b""


@pytest.mark.asyncio
async def test_websocket_send(send, receive, message_stack):
    text_string = "hello"
    text_bytes = b"hello"

    ws = WebSocketConnection(send, receive)
    await ws.send(text_string)
    await ws.send(text_bytes)

    assert len(message_stack) == 2

    message = message_stack.popleft()
    assert message["type"] == "websocket.send"
    assert message["text"] == text_string
    assert "bytes" not in message

    message = message_stack.popleft()
    assert message["type"] == "websocket.send"
    assert message["bytes"] == text_bytes
    assert "text" not in message


@pytest.mark.asyncio
async def test_websocket_text_receive(send, receive, message_stack):
    msg = {"text": "hello", "type": "websocket.receive"}
    message_stack.append(msg)

    ws = WebSocketConnection(send, receive)
    text = await ws.receive()

    assert text == msg["text"]


@pytest.mark.asyncio
async def test_websocket_bytes_receive(send, receive, message_stack):
    msg = {"bytes": b"hello", "type": "websocket.receive"}
    message_stack.append(msg)

    ws = WebSocketConnection(send, receive)
    data = await ws.receive()

    assert data == msg["bytes"]


@pytest.mark.asyncio
async def test_websocket_accept_with_no_subprotocols(
    send, receive, message_stack
):
    ws = WebSocketConnection(send, receive)
    await ws.accept()

    assert len(message_stack) == 1

    message = message_stack.popleft()
    assert message["type"] == "websocket.accept"
    assert message["subprotocol"] is None
    assert "bytes" not in message


@pytest.mark.asyncio
async def test_websocket_accept_with_subprotocol(send, receive, message_stack):
    subprotocols = ["graphql-ws"]

    ws = WebSocketConnection(send, receive, subprotocols)
    await ws.accept(subprotocols)

    assert len(message_stack) == 1

    message = message_stack.popleft()
    assert message["type"] == "websocket.accept"
    assert message["subprotocol"] == "graphql-ws"
    assert "bytes" not in message


@pytest.mark.asyncio
async def test_websocket_accept_with_multiple_subprotocols(
    send, receive, message_stack
):
    subprotocols = ["graphql-ws", "hello", "world"]

    ws = WebSocketConnection(send, receive, subprotocols)
    await ws.accept(["hello", "world"])

    assert len(message_stack) == 1

    message = message_stack.popleft()
    assert message["type"] == "websocket.accept"
    assert message["subprotocol"] == "hello"
    assert "bytes" not in message


def test_improper_websocket_connection(transport, send, receive):
    with pytest.raises(BadRequest):
        transport.get_websocket_connection()

    transport.create_websocket_connection(send, receive)
    connection = transport.get_websocket_connection()
    assert isinstance(connection, WebSocketConnection)


@pytest.mark.asyncio
async def test_request_class_regular(app):
    @app.get("/regular")
    def regular_request(request):
        return text(request.__class__.__name__)

    _, response = await app.asgi_client.get("/regular")
    assert response.body == b"Request"


@pytest.mark.asyncio
async def test_request_class_custom():
    class MyCustomRequest(Request):
        pass

    app = Sanic(name="Test", request_class=MyCustomRequest)

    @app.get("/custom")
    def custom_request(request):
        return text(request.__class__.__name__)

    _, response = await app.asgi_client.get("/custom")
    assert response.body == b"MyCustomRequest"


@pytest.mark.asyncio
async def test_cookie_customization(app):
    @app.get("/cookie")
    def get_cookie(request):
        response = text("There's a cookie up in this response")
        response.cookies["test"] = "Cookie1"
        response.cookies["test"]["httponly"] = True

        response.cookies["c2"] = "Cookie2"
        response.cookies["c2"]["httponly"] = False

        return response

    _, response = await app.asgi_client.get("/cookie")

    CookieDef = namedtuple("CookieDef", ("value", "httponly"))
    Cookie = namedtuple("Cookie", ("domain", "path", "value", "httponly"))
    cookie_map = {
        "test": CookieDef("Cookie1", True),
        "c2": CookieDef("Cookie2", False),
    }

    cookies = {
        c.name: Cookie(c.domain, c.path, c.value, "HttpOnly" in c._rest.keys())
        for c in response.cookies.jar
    }

    for name, definition in cookie_map.items():
        cookie = cookies.get(name)
        assert cookie
        assert cookie.value == definition.value
        assert cookie.domain == "mockserver.local"
        assert cookie.path == "/"
        assert cookie.httponly == definition.httponly


@pytest.mark.asyncio
async def test_content_type(app):
    @app.get("/json")
    def send_json(request):
        return json({"foo": "bar"})

    @app.get("/text")
    def send_text(request):
        return text("foobar")

    @app.get("/custom")
    def send_custom(request):
        return text("foobar", content_type="somethingelse")

    _, response = await app.asgi_client.get("/json")
    assert response.headers.get("content-type") == "application/json"

    _, response = await app.asgi_client.get("/text")
    assert response.headers.get("content-type") == "text/plain; charset=utf-8"

    _, response = await app.asgi_client.get("/custom")
    assert response.headers.get("content-type") == "somethingelse"


@pytest.mark.asyncio
async def test_request_handle_exception(app):
    @app.get("/error-prone")
    def _request(request):
        raise ServiceUnavailable(message="Service unavailable")

    _, response = await app.asgi_client.get("/wrong-path")
    assert response.status_code == 404

    _, response = await app.asgi_client.get("/error-prone")
    assert response.status_code == 503


@pytest.mark.asyncio
async def test_request_exception_suppressed_by_middleware(app):
    @app.get("/error-prone")
    def _request(request):
        raise ServiceUnavailable(message="Service unavailable")

    @app.on_request
    def forbidden(request):
        raise Forbidden(message="forbidden")

    _, response = await app.asgi_client.get("/wrong-path")
    assert response.status_code == 403

    _, response = await app.asgi_client.get("/error-prone")
    assert response.status_code == 403


@pytest.mark.asyncio
async def test_signals_triggered(app):
    @app.get("/test_signals_triggered")
    async def _request(request):
        return text("test_signals_triggered")

    signals_triggered = []
    signals_expected = [
        # "http.lifecycle.begin",
        # "http.lifecycle.read_head",
        "http.lifecycle.request",
        "http.lifecycle.handle",
        "http.routing.before",
        "http.routing.after",
        "http.handler.before",
        "http.handler.after",
        "http.lifecycle.response",
        # "http.lifecycle.send",
        # "http.lifecycle.complete",
    ]

    def signal_handler(signal):
        return lambda *a, **kw: signals_triggered.append(signal)

    for signal in RESERVED_NAMESPACES["http"]:
        app.signal(signal)(signal_handler(signal))

    _, response = await app.asgi_client.get("/test_signals_triggered")
    assert response.status_code == 200
    assert response.text == "test_signals_triggered"
    assert signals_triggered == signals_expected


@pytest.mark.asyncio
async def test_asgi_serve_location(app):
    @app.get("/")
    def _request(request: Request):
        return text(request.app.serve_location)

    _, response = await app.asgi_client.get("/")
    assert response.text == "http://<ASGI>"


@pytest.mark.asyncio
async def test_error_on_lifespan_exception_start(app, caplog):
    @app.before_server_start
    async def before_server_start(_):
        1 / 0

    recv = AsyncMock(
        side_effect=[
            {"type": "lifespan.startup"},
            {"type": "lifespan.shutdown"},
        ]
    )
    send = AsyncMock()
    app.asgi = True

    lifespan = Lifespan(app, {"type": "lifespan"}, recv, send)
    with caplog.at_level(logging.ERROR):
        await lifespan()

    send.assert_has_calls(
        [
            call(
                {
                    "type": "lifespan.startup.failed",
                    "message": "division by zero",
                }
            )
        ]
    )


@pytest.mark.asyncio
async def test_error_on_lifespan_exception_stop(app: Sanic):
    @app.before_server_stop
    async def before_server_stop(_):
        1 / 0

    recv = AsyncMock(
        side_effect=[
            {"type": "lifespan.startup"},
            {"type": "lifespan.shutdown"},
        ]
    )
    send = AsyncMock()
    app.asgi = True
    await app._startup()

    lifespan = Lifespan(app, {"type": "lifespan"}, recv, send)
    await lifespan()

    send.assert_has_calls(
        [
            call(
                {
                    "type": "lifespan.shutdown.failed",
                    "message": "division by zero",
                }
            )
        ]
    )


@pytest.mark.asyncio
async def test_asgi_headers_decoding(app: Sanic, monkeypatch: MonkeyPatch):
    @app.get("/")
    def handler(request: Request):
        return text("")

    headers_init = Headers.__init__

    def mocked_headers_init(self, *args, **kwargs):
        if "encoding" in kwargs:
            kwargs.pop("encoding")
        headers_init(self, encoding="utf-8", *args, **kwargs)

    monkeypatch.setattr(Headers, "__init__", mocked_headers_init)

    message = "Header names can only contain US-ASCII characters"
    with pytest.raises(BadRequest, match=message):
        _, response = await app.asgi_client.get("/", headers={"😂": "😅"})

    _, response = await app.asgi_client.get("/", headers={"Test-Header": "😅"})
    assert response.status_code == 200


@pytest.mark.asyncio
async def test_asgi_url_decoding(app):
    @app.get("/dir/<name>", unquote=True)
    def _request(request: Request, name):
        return text(name)

    # 2F should not become a path separator (unquoted later)
    _, response = await app.asgi_client.get("/dir/some%2Fpath")
    assert response.text == "some/path"

    _, response = await app.asgi_client.get("/dir/some%F0%9F%98%80path")
    assert response.text == "some😀path"



import asyncio
import inspect
import os
import time

from collections import namedtuple
from datetime import datetime, timedelta
from email.utils import formatdate, parsedate_to_datetime
from logging import ERROR, LogRecord
from mimetypes import guess_type
from pathlib import Path
from random import choice
from typing import Callable, List, Union
from urllib.parse import unquote

import pytest

from aiofiles import os as async_os
from pytest import LogCaptureFixture

from sanic import Request, Sanic
from sanic.compat import Header
from sanic.constants import DEFAULT_HTTP_CONTENT_TYPE
from sanic.cookies import CookieJar
from sanic.response import (
    HTTPResponse,
    ResponseStream,
    empty,
    file,
    file_stream,
    json,
    raw,
    text,
)


JSON_DATA = {"ok": True}


@pytest.mark.filterwarnings("ignore:Types other than str will be")
def test_response_body_not_a_string(app):
    """Test when a response body sent from the application is not a string"""
    random_num = choice(range(1000))

    @app.route("/hello")
    async def hello_route(request: Request):
        return text(random_num)

    request, response = app.test_client.get("/hello")
    assert response.status == 500
    assert b"Internal Server Error" in response.body


async def sample_streaming_fn(request, response=None):
    if not response:
        response = await request.respond(content_type="text/csv")
    await response.send("foo,")
    await asyncio.sleep(0.001)
    await response.send("bar")
    await response.eof()


def test_method_not_allowed():
    app = Sanic("app")

    @app.get("/")
    async def test_get(request: Request):
        return response.json({"hello": "world"})

    request, response = app.test_client.head("/")
    assert set(response.headers["Allow"].split(", ")) == {
        "GET",
    }

    request, response = app.test_client.post("/")
    assert set(response.headers["Allow"].split(", ")) == {
        "GET",
    }

    app.router.reset()

    @app.post("/")
    async def test_post(request: Request):
        return response.json({"hello": "world"})

    request, response = app.test_client.head("/")
    assert response.status == 405
    assert set(response.headers["Allow"].split(", ")) == {
        "GET",
        "POST",
    }
    assert response.headers["Content-Length"] == "0"

    request, response = app.test_client.patch("/")
    assert response.status == 405
    assert set(response.headers["Allow"].split(", ")) == {
        "GET",
        "POST",
    }
    assert response.headers["Content-Length"] == "0"


def test_response_header(app):
    @app.get("/")
    async def test(request: Request):
        return json({"ok": True}, headers={"CONTENT-TYPE": "application/json"})

    request, response = app.test_client.get("/")
    for key, value in {
        "connection": "keep-alive",
        "content-length": "11",
        "content-type": "application/json",
    }.items():
        assert response.headers[key] == value


def test_response_content_length(app):
    @app.get("/response_with_space")
    async def response_with_space(request: Request):
        return json(
            {"message": "Data", "details": "Some Details"},
            headers={"CONTENT-TYPE": "application/json"},
        )

    @app.get("/response_without_space")
    async def response_without_space(request: Request):
        return json(
            {"message": "Data", "details": "Some Details"},
            headers={"CONTENT-TYPE": "application/json"},
        )

    _, response = app.test_client.get("/response_with_space")
    content_length_for_response_with_space = response.headers.get(
        "Content-Length"
    )

    _, response = app.test_client.get("/response_without_space")
    content_length_for_response_without_space = response.headers.get(
        "Content-Length"
    )

    assert (
        content_length_for_response_with_space
        == content_length_for_response_without_space
    )

    assert content_length_for_response_with_space == "43"


def test_response_content_length_with_different_data_types(app):
    @app.get("/")
    async def get_data_with_different_types(request: Request):
        # Indentation issues in the Response is intentional. Please do not fix
        return json(
            {"bool": True, "none": None, "string": "string", "number": -1},
            headers={"CONTENT-TYPE": "application/json"},
        )

    _, response = app.test_client.get("/")
    assert response.headers.get("Content-Length") == "55"


@pytest.fixture
def json_app(app):
    @app.route("/")
    async def test(request: Request):
        return json(JSON_DATA)

    @app.get("/no-content")
    async def no_content_handler(request: Request):
        return json(JSON_DATA, status=204)

    @app.get("/no-content/unmodified")
    async def no_content_unmodified_handler(request: Request):
        return json(None, status=304)

    @app.get("/unmodified")
    async def unmodified_handler(request: Request):
        return json(JSON_DATA, status=304)

    @app.get("/precondition")
    async def precondition_handler(request: Request):
        return json(JSON_DATA, status=412)

    @app.delete("/")
    async def delete_handler(request: Request):
        return json(None, status=204)

    return app


def test_json_response(json_app):
    from sanic.response import json_dumps

    request, response = json_app.test_client.get("/")
    assert response.status == 200
    assert response.text == json_dumps(JSON_DATA)
    assert response.json == JSON_DATA

    request, response = json_app.test_client.get("/precondition")
    assert response.status == 412
    assert response.json == JSON_DATA


def test_no_content(json_app):
    request, response = json_app.test_client.get("/no-content")
    assert response.status == 204
    assert response.text == ""
    assert "Content-Length" not in response.headers

    request, response = json_app.test_client.get("/no-content/unmodified")
    assert response.status == 304
    assert response.text == ""
    assert "Content-Length" not in response.headers
    assert "Content-Type" not in response.headers

    request, response = json_app.test_client.get("/unmodified")
    assert response.status == 304
    assert response.text == ""
    assert "Content-Length" not in response.headers
    assert "Content-Type" not in response.headers

    request, response = json_app.test_client.delete("/")
    assert response.status == 204
    assert response.text == ""
    assert "Content-Length" not in response.headers


@pytest.fixture
def streaming_app(app):
    @app.route("/")
    async def test(request: Request):
        await sample_streaming_fn(request)

    return app


@pytest.fixture
def non_chunked_streaming_app(app):
    @app.route("/")
    async def test(request: Request):
        response = await request.respond(
            headers={"Content-Length": "7"},
            content_type="text/csv",
        )
        await sample_streaming_fn(request, response)

    return app


def test_chunked_streaming_adds_correct_headers(streaming_app):
    request, response = streaming_app.test_client.get("/")
    assert response.headers["Transfer-Encoding"] == "chunked"
    assert response.headers["Content-Type"] == "text/csv"
    # Content-Length is not allowed by HTTP/1.1 specification
    # when "Transfer-Encoding: chunked" is used
    assert "Content-Length" not in response.headers


def test_chunked_streaming_returns_correct_content(streaming_app):
    request, response = streaming_app.test_client.get("/")
    assert response.text == "foo,bar"


@pytest.mark.asyncio
async def test_chunked_streaming_returns_correct_content_asgi(streaming_app):
    request, response = await streaming_app.asgi_client.get("/")
    assert response.body == b"foo,bar"


def test_non_chunked_streaming_adds_correct_headers(non_chunked_streaming_app):
    request, response = non_chunked_streaming_app.test_client.get("/")

    assert "Transfer-Encoding" not in response.headers
    assert response.headers["Content-Type"] == "text/csv"
    assert response.headers["Content-Length"] == "7"


@pytest.mark.asyncio
async def test_non_chunked_streaming_adds_correct_headers_asgi(
    non_chunked_streaming_app,
):
    request, response = await non_chunked_streaming_app.asgi_client.get("/")
    assert "Transfer-Encoding" not in response.headers
    assert response.headers["Content-Type"] == "text/csv"
    assert response.headers["Content-Length"] == "7"


def test_non_chunked_streaming_returns_correct_content(
    non_chunked_streaming_app,
):
    request, response = non_chunked_streaming_app.test_client.get("/")
    assert response.text == "foo,bar"


def test_stream_response_with_cookies(app):
    @app.route("/")
    async def test(request: Request):
        headers = Header()
        cookies = CookieJar(headers)
        cookies["test"] = "modified"
        cookies["test"] = "pass"
        response = await request.respond(
            content_type="text/csv", headers=headers
        )

        await response.send("foo,")
        await asyncio.sleep(0.001)
        await response.send("bar")

    request, response = app.test_client.get("/")
    assert response.cookies["test"] == "pass"


def test_stream_response_without_cookies(app):
    @app.route("/")
    async def test(request: Request):
        await sample_streaming_fn(request)

    request, response = app.test_client.get("/")
    assert response.cookies == {}


@pytest.fixture
def static_file_directory():
    """The static directory to serve"""
    current_file = inspect.getfile(inspect.currentframe())
    current_directory = os.path.dirname(os.path.abspath(current_file))
    static_directory = os.path.join(current_directory, "static")
    return static_directory


def path_str_to_path_obj(static_file_directory: Union[Path, str]):
    if isinstance(static_file_directory, str):
        static_file_directory = Path(static_file_directory)
    return static_file_directory


def get_file_content(static_file_directory: Union[Path, str], file_name: str):
    """The content of the static file to check"""
    static_file_directory = path_str_to_path_obj(static_file_directory)
    with open(static_file_directory / file_name, "rb") as file:
        return file.read()


def get_file_last_modified_timestamp(
    static_file_directory: Union[Path, str], file_name: str
):
    """The content of the static file to check"""
    static_file_directory = path_str_to_path_obj(static_file_directory)
    return (static_file_directory / file_name).stat().st_mtime


@pytest.mark.parametrize(
    "file_name", ["test.file", "decode me.txt", "python.png"]
)
@pytest.mark.parametrize("status", [200, 401])
def test_file_response(app: Sanic, file_name, static_file_directory, status):
    @app.route("/files/<filename>", methods=["GET"])
    def file_route(request, filename):
        file_path = os.path.join(static_file_directory, filename)
        file_path = os.path.abspath(unquote(file_path))
        return file(
            file_path,
            status=status,
            mime_type=guess_type(file_path)[0] or "text/plain",
        )

    request, response = app.test_client.get(f"/files/{file_name}")
    assert response.status == status
    assert response.body == get_file_content(static_file_directory, file_name)
    assert "Content-Disposition" not in response.headers


@pytest.mark.parametrize(
    "source,dest",
    [
        ("test.file", "my_file.txt"),
        ("decode me.txt", "readme.md"),
        ("python.png", "logo.png"),
    ],
)
def test_file_response_custom_filename(
    app: Sanic, source, dest, static_file_directory
):
    @app.route("/files/<filename>", methods=["GET"])
    def file_route(request, filename):
        file_path = os.path.join(static_file_directory, filename)
        file_path = os.path.abspath(unquote(file_path))
        return file(file_path, filename=dest)

    request, response = app.test_client.get(f"/files/{source}")
    assert response.status == 200
    assert response.body == get_file_content(static_file_directory, source)
    assert (
        response.headers["Content-Disposition"]
        == f'attachment; filename="{dest}"'
    )


@pytest.mark.parametrize("file_name", ["test.file", "decode me.txt"])
def test_file_head_response(app: Sanic, file_name, static_file_directory):
    @app.route("/files/<filename>", methods=["GET", "HEAD"])
    async def file_route(request, filename):
        file_path = os.path.join(static_file_directory, filename)
        file_path = os.path.abspath(unquote(file_path))
        stats = await async_os.stat(file_path)
        headers = {}
        headers["Accept-Ranges"] = "bytes"
        headers["Content-Length"] = str(stats.st_size)
        if request.method == "HEAD":
            return HTTPResponse(
                headers=headers,
                content_type=guess_type(file_path)[0] or "text/plain",
            )
        else:
            return file(
                file_path,
                headers=headers,
                mime_type=guess_type(file_path)[0] or "text/plain",
            )

    request, response = app.test_client.head(f"/files/{file_name}")
    assert response.status == 200
    assert "Accept-Ranges" in response.headers
    assert "Content-Length" in response.headers
    assert int(response.headers["Content-Length"]) == len(
        get_file_content(static_file_directory, file_name)
    )


@pytest.mark.parametrize(
    "file_name", ["test.file", "decode me.txt", "python.png"]
)
def test_file_stream_response(app: Sanic, file_name, static_file_directory):
    @app.route("/files/<filename>", methods=["GET"])
    def file_route(request, filename):
        file_path = os.path.join(static_file_directory, filename)
        file_path = os.path.abspath(unquote(file_path))
        return file_stream(
            file_path,
            chunk_size=32,
            mime_type=guess_type(file_path)[0] or "text/plain",
        )

    request, response = app.test_client.get(f"/files/{file_name}")
    assert response.status == 200
    assert response.body == get_file_content(static_file_directory, file_name)
    assert "Content-Disposition" not in response.headers


@pytest.mark.parametrize(
    "source,dest",
    [
        ("test.file", "my_file.txt"),
        ("decode me.txt", "readme.md"),
        ("python.png", "logo.png"),
    ],
)
def test_file_stream_response_custom_filename(
    app: Sanic, source, dest, static_file_directory
):
    @app.route("/files/<filename>", methods=["GET"])
    def file_route(request, filename):
        file_path = os.path.join(static_file_directory, filename)
        file_path = os.path.abspath(unquote(file_path))
        return file_stream(file_path, chunk_size=32, filename=dest)

    request, response = app.test_client.get(f"/files/{source}")
    assert response.status == 200
    assert response.body == get_file_content(static_file_directory, source)
    assert (
        response.headers["Content-Disposition"]
        == f'attachment; filename="{dest}"'
    )


@pytest.mark.parametrize("file_name", ["test.file", "decode me.txt"])
def test_file_stream_head_response(
    app: Sanic, file_name, static_file_directory
):
    @app.route("/files/<filename>", methods=["GET", "HEAD"])
    async def file_route(request, filename):
        file_path = os.path.join(static_file_directory, filename)
        file_path = os.path.abspath(unquote(file_path))
        headers = {}
        headers["Accept-Ranges"] = "bytes"
        if request.method == "HEAD":
            # Return a normal HTTPResponse, not a
            # StreamingHTTPResponse for a HEAD request
            stats = await async_os.stat(file_path)
            headers["Content-Length"] = str(stats.st_size)
            return HTTPResponse(
                headers=headers,
                content_type=guess_type(file_path)[0] or "text/plain",
            )
        else:
            return file_stream(
                file_path,
                chunk_size=32,
                headers=headers,
                mime_type=guess_type(file_path)[0] or "text/plain",
            )

    request, response = app.test_client.head(f"/files/{file_name}")
    assert response.status == 200
    # A HEAD request should never be streamed/chunked.
    if "Transfer-Encoding" in response.headers:
        assert response.headers["Transfer-Encoding"] != "chunked"
    assert "Accept-Ranges" in response.headers
    # A HEAD request should get the Content-Length too
    assert "Content-Length" in response.headers
    assert int(response.headers["Content-Length"]) == len(
        get_file_content(static_file_directory, file_name)
    )


@pytest.mark.parametrize(
    "file_name", ["test.file", "decode me.txt", "python.png"]
)
@pytest.mark.parametrize(
    "size,start,end", [(1024, 0, 1024), (4096, 1024, 8192)]
)
def test_file_stream_response_range(
    app: Sanic, file_name, static_file_directory, size, start, end
):
    Range = namedtuple("Range", ["size", "start", "end", "total"])
    total = len(get_file_content(static_file_directory, file_name))
    range = Range(size=size, start=start, end=end, total=total)

    @app.route("/files/<filename>", methods=["GET"])
    def file_route(request, filename):
        file_path = os.path.join(static_file_directory, filename)
        file_path = os.path.abspath(unquote(file_path))
        return file_stream(
            file_path,
            chunk_size=32,
            mime_type=guess_type(file_path)[0] or "text/plain",
            _range=range,
        )

    request, response = app.test_client.get(f"/files/{file_name}")
    assert response.status == 206
    assert "Content-Range" in response.headers
    assert (
        response.headers["Content-Range"]
        == f"bytes {range.start}-{range.end}/{range.total}"
    )


def test_raw_response(app):
    @app.get("/test")
    def handler(request: Request):
        return raw(b"raw_response")

    request, response = app.test_client.get("/test")
    assert response.content_type == DEFAULT_HTTP_CONTENT_TYPE
    assert response.body == b"raw_response"


def test_empty_response(app):
    @app.get("/test")
    def handler(request: Request):
        return empty()

    request, response = app.test_client.get("/test")
    assert response.content_type is None
    assert response.body == b""


def test_direct_response_stream(app: Sanic):
    @app.route("/")
    async def test(request: Request):
        response = await request.respond(content_type="text/csv")
        await response.send("foo,")
        await response.send("bar")
        await response.eof()

    _, response = app.test_client.get("/")
    assert response.text == "foo,bar"
    assert response.headers["Transfer-Encoding"] == "chunked"
    assert response.headers["Content-Type"] == "text/csv"
    assert "Content-Length" not in response.headers


@pytest.mark.asyncio
async def test_direct_response_stream_asgi(app: Sanic):
    @app.route("/")
    async def test(request: Request):
        response = await request.respond(content_type="text/csv")
        await response.send("foo,")
        await response.send("bar")
        await response.eof()

    _, response = await app.asgi_client.get("/")
    assert response.text == "foo,bar"
    assert response.headers["Content-Type"] == "text/csv"
    assert "Content-Length" not in response.headers


def test_multiple_responses(
    app: Sanic,
    caplog: LogCaptureFixture,
    message_in_records: Callable[[List[LogRecord], str], bool],
):
    @app.route("/1")
    async def handler1(request: Request):
        response = await request.respond()
        await response.send("foo")
        response = await request.respond()

    @app.route("/2")
    async def handler2(request: Request):
        response = await request.respond()
        response = await request.respond()
        await response.send("foo")

    @app.get("/3")
    async def handler3(request: Request):
        response = await request.respond()
        await response.send("foo,")
        response = await request.respond()
        await response.send("bar")

    @app.get("/4")
    async def handler4(request: Request):
        await request.respond(headers={"one": "one"})
        return json({"foo": "bar"}, headers={"one": "two"})

    @app.get("/5")
    async def handler5(request: Request):
        response = await request.respond(headers={"one": "one"})
        await response.send("foo")
        return json({"foo": "bar"}, headers={"one": "two"})

    @app.get("/6")
    async def handler6(request: Request):
        response = await request.respond(headers={"one": "one"})
        await response.send("foo, ")
        json_response = json({"foo": "bar"}, headers={"one": "two"})
        await response.send("bar")
        return json_response

    error_msg0 = "Second respond call is not allowed."

    error_msg1 = (
        "The error response will not be sent to the client for the following "
        'exception:"Second respond call is not allowed.". A previous '
        "response has at least partially been sent."
    )

    error_msg2 = (
        "The response object returned by the route handler "
        "will not be sent to client. The request has already "
        "been responded to."
    )

    with caplog.at_level(ERROR):
        _, response = app.test_client.get("/1")
        assert response.status == 200
        assert message_in_records(caplog.records, error_msg0)
        assert message_in_records(caplog.records, error_msg1)

    with caplog.at_level(ERROR):
        _, response = app.test_client.get("/2")
        assert response.status == 500
        assert "500 — Internal Server Error" in response.text

    with caplog.at_level(ERROR):
        _, response = app.test_client.get("/3")
        assert response.status == 200
        assert "foo," in response.text
        assert message_in_records(caplog.records, error_msg0)
        assert message_in_records(caplog.records, error_msg1)

    with caplog.at_level(ERROR):
        _, response = app.test_client.get("/4")
        assert response.status == 200
        assert "foo" not in response.text
        assert "one" in response.headers
        assert response.headers["one"] == "one"

        assert message_in_records(caplog.records, error_msg2)

    with caplog.at_level(ERROR):
        _, response = app.test_client.get("/5")
        assert response.status == 200
        assert "foo" in response.text
        assert "one" in response.headers
        assert response.headers["one"] == "one"
        assert message_in_records(caplog.records, error_msg2)

    with caplog.at_level(ERROR):
        _, response = app.test_client.get("/6")
        assert "foo, bar" in response.text
        assert "one" in response.headers
        assert response.headers["one"] == "one"
        assert message_in_records(caplog.records, error_msg2)


def test_send_response_after_eof_should_fail(
    app: Sanic,
    caplog: LogCaptureFixture,
    message_in_records: Callable[[List[LogRecord], str], bool],
):
    @app.get("/")
    async def handler(request: Request):
        response = await request.respond()
        await response.send("foo, ")
        await response.eof()
        await response.send("bar")

    error_msg1 = (
        "The error response will not be sent to the client for the following "
        'exception:"Response stream was ended, no more response '
        'data is allowed to be sent.". A previous '
        "response has at least partially been sent."
    )

    error_msg2 = (
        "Response stream was ended, no more response "
        "data is allowed to be sent."
    )

    with caplog.at_level(ERROR):
        _, response = app.test_client.get("/")
        assert "foo, " in response.text
        assert message_in_records(caplog.records, error_msg1)
        assert message_in_records(caplog.records, error_msg2)


@pytest.mark.asyncio
async def test_send_response_after_eof_should_fail_asgi(
    app: Sanic,
    caplog: LogCaptureFixture,
    message_in_records: Callable[[List[LogRecord], str], bool],
):
    @app.get("/")
    async def handler(request: Request):
        response = await request.respond()
        await response.send("foo, ")
        await response.eof()
        await response.send("bar")

    error_msg1 = (
        "The error response will not be sent to the client for the "
        'following exception:"There is no request to respond to, '
        "either the response has already been sent or the request "
        'has not been received yet.". A previous response has '
        "at least partially been sent."
    )

    error_msg2 = (
        "There is no request to respond to, either the response has "
        "already been sent or the request has not been received yet."
    )

    with caplog.at_level(ERROR):
        _, response = await app.asgi_client.get("/")
        assert "foo, " in response.text
        assert message_in_records(caplog.records, error_msg1)
        assert message_in_records(caplog.records, error_msg2)


@pytest.mark.parametrize(
    "file_name", ["test.file", "decode me.txt", "python.png"]
)
def test_file_response_headers(
    app: Sanic, file_name: str, static_file_directory: str
):
    test_last_modified = datetime.now()
    test_max_age = 10
    test_expires = test_last_modified.timestamp() + test_max_age

    @app.route("/files/cached/<filename>", methods=["GET"])
    def file_route_cache(request: Request, filename: str):
        file_path = (
            Path(static_file_directory) / unquote(filename)
        ).absolute()
        return file(
            file_path, max_age=test_max_age, last_modified=test_last_modified
        )

    @app.route(
        "/files/cached_default_last_modified/<filename>", methods=["GET"]
    )
    def file_route_cache_default_last_modified(
        request: Request, filename: str
    ):
        file_path = (
            Path(static_file_directory) / unquote(filename)
        ).absolute()
        return file(file_path, max_age=test_max_age)

    @app.route("/files/no_cache/<filename>", methods=["GET"])
    def file_route_no_cache(request: Request, filename: str):
        file_path = (
            Path(static_file_directory) / unquote(filename)
        ).absolute()
        return file(file_path)

    @app.route("/files/no_store/<filename>", methods=["GET"])
    def file_route_no_store(request: Request, filename: str):
        file_path = (
            Path(static_file_directory) / unquote(filename)
        ).absolute()
        return file(file_path, no_store=True)

    _, response = app.test_client.get(f"/files/cached/{file_name}")
    assert response.body == get_file_content(static_file_directory, file_name)
    headers = response.headers
    assert (
        "cache-control" in headers
        and f"max-age={test_max_age}" in headers.get("cache-control")
        and "public" in headers.get("cache-control")
    )
    assert (
        "expires" in headers
        and headers.get("expires")[:-6]
        == formatdate(test_expires, usegmt=True)[:-6]
        # [:-6] to allow at most 1 min difference
        # It's minimal for cases like:
        # Thu, 26 May 2022 05:36:59 GMT
        # AND
        # Thu, 26 May 2022 05:37:00 GMT
    )
    assert response.status == 200
    assert "last-modified" in headers and headers.get(
        "last-modified"
    ) == formatdate(test_last_modified.timestamp(), usegmt=True)

    _, response = app.test_client.get(
        f"/files/cached_default_last_modified/{file_name}"
    )
    file_last_modified = get_file_last_modified_timestamp(
        static_file_directory, file_name
    )
    headers = response.headers
    assert "last-modified" in headers and headers.get(
        "last-modified"
    ) == formatdate(file_last_modified, usegmt=True)
    assert response.status == 200

    _, response = app.test_client.get(f"/files/no_cache/{file_name}")
    headers = response.headers
    assert "cache-control" in headers and "no-cache" == headers.get(
        "cache-control"
    )
    assert response.status == 200

    _, response = app.test_client.get(f"/files/no_store/{file_name}")
    headers = response.headers
    assert "cache-control" in headers and "no-store" == headers.get(
        "cache-control"
    )
    assert response.status == 200


def test_file_validate(app: Sanic, static_file_directory: str):
    file_name = "test_validate.txt"
    static_file_directory = Path(static_file_directory)
    file_path = static_file_directory / file_name
    file_path = file_path.absolute()
    test_max_age = 10

    with open(file_path, "w+") as f:
        f.write("foo\n")

    @app.route("/validate", methods=["GET"])
    def file_route_cache(request: Request):
        return file(
            file_path,
            request_headers=request.headers,
            max_age=test_max_age,
            validate_when_requested=True,
        )

    _, response = app.test_client.get("/validate")
    assert response.status == 200
    assert response.body == b"foo\n"
    last_modified = response.headers["Last-Modified"]

    time.sleep(1)
    with open(file_path, "a") as f:
        f.write("bar\n")
    _, response = app.test_client.get(
        "/validate", headers={"If-Modified-Since": last_modified}
    )

    assert response.status == 200
    assert response.body == b"foo\nbar\n"

    last_modified = response.headers["Last-Modified"]
    _, response = app.test_client.get(
        "/validate", headers={"if-modified-since": last_modified}
    )
    assert response.status == 304
    assert response.body == b""

    file_path.unlink()


@pytest.mark.parametrize(
    "file_name", ["test.file", "decode me.txt", "python.png"]
)
def test_file_validating_invalid_header(
    app: Sanic, file_name: str, static_file_directory: str
):
    @app.route("/files/<filename>", methods=["GET"])
    def file_route(request: Request, filename: str):
        handler_file_path = (
            Path(static_file_directory) / unquote(filename)
        ).absolute()

        return file(
            handler_file_path,
            request_headers=request.headers,
            validate_when_requested=True,
        )

    _, response = app.test_client.get(f"/files/{file_name}")
    assert response.status == 200
    assert response.body == get_file_content(static_file_directory, file_name)

    _, response = app.test_client.get(
        f"/files/{file_name}", headers={"if-modified-since": "invalid-value"}
    )
    assert response.status == 200
    assert response.body == get_file_content(static_file_directory, file_name)

    _, response = app.test_client.get(
        f"/files/{file_name}", headers={"if-modified-since": ""}
    )
    assert response.status == 200
    assert response.body == get_file_content(static_file_directory, file_name)


@pytest.mark.parametrize(
    "file_name", ["test.file", "decode me.txt", "python.png"]
)
def test_file_validating_304_response_file_route(
    app: Sanic, file_name: str, static_file_directory: str
):
    @app.route("/files/<filename>", methods=["GET"])
    def file_route(request: Request, filename: str):
        handler_file_path = (
            Path(static_file_directory) / unquote(filename)
        ).absolute()

        return file(
            handler_file_path,
            request_headers=request.headers,
            validate_when_requested=True,
        )

    _, response = app.test_client.get(f"/files/{file_name}")
    assert response.status == 200
    assert response.body == get_file_content(static_file_directory, file_name)

    _, response = app.test_client.get(
        f"/files/{file_name}",
        headers={"if-modified-since": response.headers["Last-Modified"]},
    )
    assert response.status == 304
    assert response.body == b""


@pytest.mark.parametrize(
    "file_name", ["test.file", "decode me.txt", "python.png"]
)
def test_file_validating_304_response(
    app: Sanic, file_name: str, static_file_directory: str
):
    app.static("static", Path(static_file_directory) / file_name)

    _, response = app.test_client.get("/static")
    assert response.status == 200
    assert response.body == get_file_content(static_file_directory, file_name)
    last_modified = parsedate_to_datetime(response.headers["Last-Modified"])
    last_modified += timedelta(seconds=1)
    _, response = app.test_client.get(
        "/static",
        headers={
            "if-modified-since": formatdate(
                last_modified.timestamp(), usegmt=True
            )
        },
    )
    assert response.status == 304
    assert response.body == b""


def test_stream_response_with_default_headers(app: Sanic):
    async def sample_streaming_fn(response_):
        await response_.write("foo")

    @app.route("/")
    async def test(request: Request):
        return ResponseStream(sample_streaming_fn, content_type="text/csv")

    _, response = app.test_client.get("/")
    assert response.text == "foo"
    assert response.headers["Transfer-Encoding"] == "chunked"
    assert response.headers["Content-Type"] == "text/csv"



from uuid import UUID

import pytest

from sanic import json


@pytest.mark.parametrize(
    "idx,path,expectation",
    (
        (0, "/abc", "str"),
        (1, "/123", "int"),
        (2, "/123.5", "float"),
        (3, "/8af729fe-2b94-4a95-a168-c07068568429", "UUID"),
    ),
)
def test_annotated_handlers(app, idx, path, expectation):
    def build_response(num, foo):
        return json({"num": num, "type": type(foo).__name__})

    @app.get("/<foo>")
    def handler0(_, foo: str):
        return build_response(0, foo)

    @app.get("/<foo>")
    def handler1(_, foo: int):
        return build_response(1, foo)

    @app.get("/<foo>")
    def handler2(_, foo: float):
        return build_response(2, foo)

    @app.get("/<foo>")
    def handler3(_, foo: UUID):
        return build_response(3, foo)

    _, response = app.test_client.get(path)
    assert response.json["num"] == idx
    assert response.json["type"] == expectation



import os
import sys
import time

from contextlib import contextmanager
from queue import Queue
from threading import Thread


if os.name == "nt":  # noqa
    import ctypes  # noqa

    class _CursorInfo(ctypes.Structure):
        _fields_ = [("size", ctypes.c_int), ("visible", ctypes.c_byte)]


class Spinner:  # noqa
    """Spinner class to show a loading spinner in the terminal.

    Used internally by the `loading` context manager.
    """

    def __init__(self, message: str) -> None:
        self.message = message
        self.queue: Queue[int] = Queue()
        self.spinner = self.cursor()
        self.thread = Thread(target=self.run)

    def start(self):
        self.queue.put(1)
        self.thread.start()
        self.hide()

    def run(self):
        while self.queue.get():
            output = f"\r{self.message} [{next(self.spinner)}]"
            sys.stdout.write(output)
            sys.stdout.flush()
            time.sleep(0.1)
            self.queue.put(1)

    def stop(self):
        self.queue.put(0)
        self.thread.join()
        self.show()

    @staticmethod
    def cursor():
        while True:
            for cursor in "|/-\\":
                yield cursor

    @staticmethod
    def hide():
        if os.name == "nt":
            ci = _CursorInfo()
            handle = ctypes.windll.kernel32.GetStdHandle(-11)
            ctypes.windll.kernel32.GetConsoleCursorInfo(
                handle, ctypes.byref(ci)
            )
            ci.visible = False
            ctypes.windll.kernel32.SetConsoleCursorInfo(
                handle, ctypes.byref(ci)
            )
        elif os.name == "posix":
            sys.stdout.write("\033[?25l")
            sys.stdout.flush()

    @staticmethod
    def show():
        if os.name == "nt":
            ci = _CursorInfo()
            handle = ctypes.windll.kernel32.GetStdHandle(-11)
            ctypes.windll.kernel32.GetConsoleCursorInfo(
                handle, ctypes.byref(ci)
            )
            ci.visible = True
            ctypes.windll.kernel32.SetConsoleCursorInfo(
                handle, ctypes.byref(ci)
            )
        elif os.name == "posix":
            sys.stdout.write("\033[?25h")
            sys.stdout.flush()


@contextmanager
def loading(message: str = "Loading"):  # noqa
    spinner = Spinner(message)
    spinner.start()
    yield
    spinner.stop()
